"""This module contains the context managers for LLM API calls."""

from __future__ import annotations

import threading
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from types import TracebackType
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast, overload

from pydantic import BaseModel
from typing_extensions import TypedDict

from ..core.base import BaseTool, BaseType, CommonCallParams
from ..core.base.stream_config import StreamConfig
from ..core.base.types import LocalProvider, Provider

_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType | Enum)

if TYPE_CHECKING:
    from ..core.anthropic import AnthropicCallParams
    from ..core.azure import AzureCallParams
    from ..core.bedrock import BedrockCallParams
    from ..core.cohere import CohereCallParams
    from ..core.gemini import GeminiCallParams
    from ..core.google import GoogleCallParams
    from ..core.groq import GroqCallParams
    from ..core.litellm import LiteLLMCallParams
    from ..core.mistral import MistralCallParams
    from ..core.openai import OpenAICallParams
    from ..core.vertex import VertexCallParams
    from ..core.xai import XAICallParams
else:
    AnthropicCallParams = AzureCallParams = BedrockCallParams = CohereCallParams = (
        GeminiCallParams
    ) = GoogleCallParams = GroqCallParams = LiteLLMCallParams = MistralCallParams = (
        OpenAICallParams
    ) = VertexCallParams = XAICallParams = None


class CallArgs(TypedDict):
    """TypedDict for call arguments."""

    provider: Provider | LocalProvider
    model: str
    stream: bool | StreamConfig
    tools: list[type[BaseTool] | Callable] | None
    response_model: type[BaseModel] | type[BaseType] | type[Enum] | None
    output_parser: Callable | None
    json_mode: bool
    client: Any | None
    call_params: CommonCallParams | Any | None


# We use a thread-local variable to store the current context, so that it's thread-safe
_current_context_local = threading.local()


@dataclass
class LLMContext:
    """Context for LLM API calls.

    This class is used to store the context for LLM API calls, including both
    setting overrides (provider, model, client, call_params) and
    structural overrides (stream, tools, response_model, etc.).
    """

    provider: Provider | LocalProvider | None = None
    model: str | None = None
    stream: bool | StreamConfig | None = None
    tools: list[type[BaseTool] | Callable] | None = None
    response_model: type[BaseModel] | type[BaseType] | type[Enum] | None = None
    output_parser: Callable | None = None
    json_mode: bool | None = None
    client: Any | None = None
    call_params: CommonCallParams | Any | None = None

    def __enter__(self) -> LLMContext:
        _current_context_local.context = self
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        if hasattr(_current_context_local, "context"):
            del _current_context_local.context
        return False  # Don't suppress exceptions


def get_current_context() -> LLMContext | None:
    """Get the current context for LLM API calls.

    Returns:
        The current context, or None if there is no context.
    """
    if hasattr(_current_context_local, "context"):
        return cast(LLMContext, _current_context_local.context)
    return None


def _context(
    *,
    provider: Provider | LocalProvider | None,
    model: str | None,
    stream: bool | StreamConfig | None = None,
    tools: list[type[BaseTool] | Callable] | None = None,
    response_model: type[_ResponseModelT] | None = None,
    output_parser: Callable | None = None,
    json_mode: bool | None = None,
    client: Any | None = None,  # noqa: ANN401
    call_params: CommonCallParams | Any | None = None,  # noqa: ANN401
) -> LLMContext:
    """Context manager for LLM API calls.

    This is an internal method that allows both setting and structural overrides
    for LLM functions.

    Unfortunately we have not yet identified a way to properly type hint this because
    providing no structural overrides means the return type is that of the original
    function. Of course, the `apply` method could pass through the return type, but
    we do not have a way to know whether it should be passthrough or not.

    For now, we use `_context` simply to implement `override` fully. The public facing
    `context` method only allows setting overrides.

    Args:
        provider: The provider to use for the LLM API call.
        model: The model to use for the LLM API call.
        stream: Whether to stream the response.
        tools: The tools to use for the LLM API call.
        response_model: The response model for the LLM API call.
        output_parser: The output parser for the LLM API call.
        json_mode: Whether to use JSON mode.
        client: The client to use for the LLM API call.
        call_params: The call parameters for the LLM API call.

    Returns:
        The context object that can be used to apply the context to a function.
    """
    old_context: LLMContext | None = getattr(_current_context_local, "context", None)
    if not old_context:
        return LLMContext(
            provider=provider,
            model=model,
            stream=stream,
            tools=tools,
            response_model=response_model,
            output_parser=output_parser,
            json_mode=json_mode,
            client=client,
            call_params=call_params,
        )
    else:
        # Ensure we properly set nested context settings. For example, we need to make
        # sure that calling override on an overridden function applies the context to
        # the overridden function's already overridden settings.
        return LLMContext(
            provider=provider or old_context.provider,
            model=model or old_context.model,
            stream=stream or old_context.stream,
            tools=tools or old_context.tools,
            response_model=response_model or old_context.response_model,
            output_parser=output_parser or old_context.output_parser,
            json_mode=json_mode or old_context.json_mode,
            client=client or old_context.client,
            call_params=call_params or old_context.call_params,
        )


def apply_context_overrides_to_call_args(
    call_args: CallArgs, context_override: LLMContext | None = None
) -> CallArgs:
    """Apply any active context overrides to the call arguments.

    Args:
        call_args: The original call arguments.
        context_override: Optional explicit context to use instead of the current thread context.

    Returns:
        The call arguments with any context overrides applied.
    """
    context = context_override or get_current_context()
    if not context:
        return call_args

    # Create a new dict with the original args
    overridden_args = CallArgs(call_args)

    # If any structural overrides are set, we have to force all others to take their
    # default values so the type hints match.
    if context.stream or context.response_model or context.output_parser:
        overridden_args["stream"] = False
        overridden_args["response_model"] = None
        overridden_args["output_parser"] = None
    if context.response_model:
        overridden_args["tools"] = None

    # Apply context overrides
    if context.provider is not None:
        overridden_args["provider"] = context.provider
    if context.model is not None:
        overridden_args["model"] = context.model
    if context.stream is not None:
        overridden_args["stream"] = context.stream
    if context.tools is not None:
        overridden_args["tools"] = context.tools
    if context.response_model is not None:
        overridden_args["response_model"] = context.response_model
    if context.output_parser is not None:
        overridden_args["output_parser"] = context.output_parser
    if context.json_mode is not None:
        overridden_args["json_mode"] = context.json_mode
    if context.client is not None:
        overridden_args["client"] = context.client
    if context.call_params is not None:
        overridden_args["call_params"] = context.call_params

    return overridden_args


@overload
def context(
    *,
    provider: Literal["anthropic"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | AnthropicCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["azure"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | AzureCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["bedrock"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | BedrockCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["cohere"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | CohereCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["gemini"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | GeminiCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["google"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | GoogleCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["groq"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | GroqCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["litellm"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | LiteLLMCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["mistral"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | MistralCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["openai"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | OpenAICallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["vertex"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | VertexCallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


@overload
def context(
    *,
    provider: Literal["xai"],
    model: str,
    client: Any = None,  # noqa: ANN401
    call_params: CommonCallParams | XAICallParams | None = None,  # noqa: ANN401
) -> LLMContext: ...


def context(
    *,
    provider: Provider | LocalProvider,
    model: str,
    client: Any | None = None,
    call_params: CommonCallParams | Any | None = None,  # noqa: ANN401
) -> LLMContext:
    """Context manager for LLM API calls.

    This method only allows setting overrides (provider, model, client, call_params)
    and does not allow structural overrides (stream, tools, response_model, etc.).

    Example:
        ```python
        @llm.call(provider="openai", model="gpt-4o-mini")
        def recommend_book(genre: str) -> str:
            return f"Recommend a {genre} book"

        # Override the model for a specific call
        with llm.context(provider="anthropic", model="claude-3-5-sonnet-20240620") as ctx:
            response = recommend_book("fantasy")  # Uses claude-3-5-sonnet
        ```

    Args:
        provider: The provider to use for the LLM API call.
        model: The model to use for the LLM API call.
        client: The client to use for the LLM API call.
        call_params: The call parameters for the LLM API call.

    Yields:
        The context object.
    """
    if (provider and not model) or (model and not provider):
        raise ValueError(
            "Provider and model must both be specified if either is specified."
        )

    return _context(
        provider=provider, model=model, client=client, call_params=call_params
    )



"""The `fallback` module provides a fallback retry strategy."""

import inspect
from collections.abc import Callable, Coroutine
from typing import Any, ParamSpec, Protocol, TypeVar, overload

from typing_extensions import NotRequired, Required, TypedDict

from .. import llm
from ..core.base import CommonCallParams
from ..core.base.types import Provider

_P = ParamSpec("_P")
_R = TypeVar("_R")


class FallbackDecorator(Protocol):
    @overload
    def __call__(
        self, fn: Callable[_P, Coroutine[_R, Any, Any]]
    ) -> Callable[_P, Coroutine[_R, Any, Any]]: ...

    @overload
    def __call__(self, fn: Callable[_P, _R]) -> Callable[_P, _R]: ...

    def __call__(
        self,
        fn: Callable[_P, _R] | Callable[_P, Coroutine[_R, Any, Any]],
    ) -> Callable[_P, _R] | Callable[_P, Coroutine[_R, Any, Any]]: ...


class Fallback(TypedDict):
    """The override arguments to use for this fallback attempt."""

    catch: Required[type[Exception] | tuple[type[Exception]]]
    provider: Required[Provider]
    model: Required[str]
    call_params: NotRequired[CommonCallParams]
    client: NotRequired[Any]


class FallbackError(Exception):
    """An error raised when all fallbacks fail."""


def fallback(
    catch: type[Exception] | tuple[type[Exception]],
    fallbacks: list[Fallback],
) -> FallbackDecorator:
    """A decorator that retries the function call with a fallback strategy.

    This must use the provider-agnostic `llm.call` decorator.

    Args:
        catch: The exception(s) to catch for the original call.
        backups: The list of backup providers to try in order. Each backup provider
            is a tuple of the provider name, the model name, and the call params.
            The call params may be `None` if no change is wanted.

    Returns:
        The decorated function.

    Raises:
        FallbackError: If all fallbacks fail.
    """

    @overload
    def decorator(
        fn: Callable[_P, Coroutine[_R, Any, Any]],
    ) -> Callable[_P, Coroutine[_R, Any, Any]]: ...

    @overload
    def decorator(fn: Callable[_P, _R]) -> Callable[_P, _R]: ...

    def decorator(
        fn: Callable[_P, _R] | Callable[_P, Coroutine[_R, Any, Any]],
    ) -> Callable[_P, _R] | Callable[_P, Coroutine[_R, Any, Any]]:
        # TODO: figure out why llm call fn is not considered as coroutine at runtime
        if inspect.iscoroutinefunction(fn._original_fn):  # pyright: ignore [reportFunctionMemberAccess]

            async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                caught: list[Exception] = []
                try:
                    return await fn(*args, **kwargs)  # pyright: ignore [reportReturnType,reportGeneralTypeIssues]
                except catch as e:
                    caught.append(e)
                    for backup in fallbacks:
                        try:
                            response = await llm.override(
                                fn,
                                provider=backup["provider"],
                                model=backup["model"],
                                call_params=backup.get("call_params", None),
                                client=backup.get("client", None),
                            )(*args, **kwargs)  # pyright: ignore [reportGeneralTypeIssues]
                            response._caught = caught  # pyright: ignore [reportAttributeAccessIssue]
                            return response  # pyright: ignore [reportReturnType]
                        except backup["catch"] as be:
                            caught.append(be)
                raise FallbackError(f"All fallbacks failed:\n{caught}")

            return inner_async
        else:

            def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                caught: list[Exception] = []
                try:
                    return fn(*args, **kwargs)  # pyright: ignore [reportReturnType]
                except catch as e:
                    caught.append(e)
                    for backup in fallbacks:
                        try:
                            response = llm.override(
                                fn,
                                provider=backup["provider"],
                                model=backup["model"],
                                call_params=backup.get("call_params", None),
                                client=backup.get("client", None),
                            )(*args, **kwargs)
                            response._caught = caught  # pyright: ignore [reportAttributeAccessIssue]
                            return response  # pyright: ignore [reportReturnType]
                        except backup["catch"] as be:
                            caught.append(be)
                raise FallbackError(f"All fallbacks failed:\n{caught}")

            return inner

    return decorator



from collections.abc import Callable
from typing import TypeVar

from .calls import BaseCall

_BaseCallT = TypeVar("_BaseCallT", bound=BaseCall)


def tags(args: list[str]) -> Callable[[type[_BaseCallT]], type[_BaseCallT]]:
    '''A decorator for adding tags to a `BasePrompt`.

    Adding this decorator to a `BasePrompt` updates the `_tags` class attribute to the
    given value. This is useful for adding metadata to a `BasePrompt` that can be used
    for logging or filtering.

    Example:

    ```python
    from mirascope import BasePrompt, tags


    @tags(["book_recommendation", "entertainment"])
    class BookRecommendationPrompt(BasePrompt):
        prompt_template = """
        SYSTEM:
        You are the world's greatest librarian.

        USER:
        I've recently read this book: {book_title}.
        What should I read next?
        """

        book_title: [str]

    print(BookRecommendationPrompt.dump()["tags"])
    #> ['book_recommendation', 'entertainment']
    ```

    Returns:
        The decorated class with `tags` class attribute set.
    '''

    def tags_fn(model_class: type[_BaseCallT]) -> type[_BaseCallT]:
        """Updates the `tags` class attribute to the given value."""
        model_class.tags = args
        return model_class

    return tags_fn



from mirascope.llm._response_metaclass import _ResponseMetaclass


def test_response_metaclass():
    class TestClass(metaclass=_ResponseMetaclass):
        def __init__(self):
            self.value_x = 1

        @property
        def value_y(self):
            return 2

    test_instance = TestClass()
    assert test_instance.value_y == 2



from collections.abc import Sequence
from functools import cached_property
from typing import Any, ClassVar, cast
from unittest.mock import MagicMock, PropertyMock, patch

import pytest
from pydantic import computed_field

from mirascope.core.base import (
    BaseCallParams,
    BaseCallResponse,
    BaseDynamicConfig,
    BaseMessageParam,
    BaseTool,
    Metadata,
    ToolResultPart,
    Usage,
)
from mirascope.core.base._utils import BaseMessageParamConverter
from mirascope.core.base.types import CostMetadata, FinishReason
from mirascope.llm.call_response import CallResponse
from mirascope.llm.tool import Tool


class DummyCallParams(BaseCallParams):
    pass


class DummyMessageParam(BaseMessageParam):
    role: str
    content: Any


class DummyTool(Tool):
    def call(self): ...

    @property
    def model_fields(self):  # pyright: ignore [reportIncompatibleVariableOverride]
        return ["field1"]

    field1: str = "tool_field"


class DummyMessageParamConverter(BaseMessageParamConverter):
    @staticmethod
    def from_provider(message_params: list[Any]) -> list[BaseMessageParam]:
        return [
            BaseMessageParam(role=message_param.role, content=message_param.content)
            for message_param in message_params
        ]

    @staticmethod
    def to_provider(message_params: list[BaseMessageParam]) -> list[Any]: ...


class DummyProviderCallResponse(
    BaseCallResponse[
        Any,
        DummyTool,
        Any,
        BaseDynamicConfig,
        DummyMessageParam,
        DummyCallParams,
        DummyMessageParam,
        Any,
    ]
):
    _message_converter: type = DummyMessageParamConverter

    @property
    def content(self) -> str:
        return "dummy_content"

    @property
    def finish_reasons(self) -> list[str] | None:
        return ["finish"]

    @property
    def model(self) -> str | None: ...

    @property
    def id(self) -> str | None: ...

    @property
    def usage(self) -> Any: ...

    @property
    def input_tokens(self) -> int | float | None: ...

    @property
    def cached_tokens(self) -> int | float | None: ...

    @property
    def output_tokens(self) -> int | float | None: ...

    @property
    def cost(self) -> float | None: ...

    @computed_field
    @cached_property
    def message_param(self) -> Any: ...

    @cached_property
    def tools(self) -> list[DummyTool] | None:
        return [DummyTool(MagicMock(spec=BaseTool))]

    @cached_property
    def tool(self) -> DummyTool | None: ...

    @classmethod
    def tool_message_params(  # pyright: ignore [reportIncompatibleMethodOverride]
        cls, tools_and_outputs: Sequence[tuple[DummyTool, str]]
    ) -> list[Any]: ...

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        return cast(list[FinishReason], self.finish_reasons)

    @property
    def common_message_param(self):
        return BaseMessageParam(role="assistant", content="common_message")

    @property
    def common_user_message_param(self):
        return BaseMessageParam(role="user", content="common_user_message")

    @property
    def common_usage(self) -> Usage | None:
        return Usage(input_tokens=1, cached_tokens=1, output_tokens=1, total_tokens=2)

    def common_construct_call_response(self): ...

    def common_construct_message_param(
        self, tool_calls: list[Any] | None, content: str | None
    ): ...

    @property
    def cost_metadata(self) -> CostMetadata: ...


@pytest.fixture
def dummy_call_response_instance():
    dummy_response = DummyProviderCallResponse(
        metadata=Metadata(),
        response={},
        tool_types=None,
        prompt_template=None,
        fn_args={},
        dynamic_config={},
        messages=[DummyMessageParam(role="assistant", content="message")],
        call_params=DummyCallParams(),
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    return CallResponse(response=dummy_response)  # pyright: ignore [reportAbstractUsage,reportArgumentType]


def test_call_response(dummy_call_response_instance):
    assert dummy_call_response_instance.finish_reasons == ["finish"]
    assert isinstance(dummy_call_response_instance.message_param, BaseMessageParam)
    message_param = dummy_call_response_instance.message_param
    assert message_param.role == "assistant"
    assert dummy_call_response_instance.tools is not None
    assert dummy_call_response_instance.tool is not None
    assert str(dummy_call_response_instance) == "dummy_content"
    assert dummy_call_response_instance._response.common_finish_reasons == ["finish"]
    assert dummy_call_response_instance.usage == Usage(
        input_tokens=1, cached_tokens=1, output_tokens=1, total_tokens=2
    )
    assert dummy_call_response_instance.common_messages == [
        BaseMessageParam(role="assistant", content="message")
    ]
    assert dummy_call_response_instance.messages == [
        BaseMessageParam(role="assistant", content="message")
    ]


def test_call_response_attribute_fallback_on_instance(dummy_call_response_instance):
    dummy_call_response_instance.custom_attr = "custom_value"  # pyright: ignore [reportAttributeAccessIssue]
    assert dummy_call_response_instance.custom_attr == "custom_value"


def test_tool_message_params_various_tool_call_ids_with_annotations():
    class ToolCallWithID:
        id = "tool_call_with_id"

    class ToolWithID(BaseTool):
        tool_call: ClassVar[Any] = ToolCallWithID()

        def call(self): ...

        field1: str = "tool_field"

    class ToolNoCall(BaseTool):  # pyright: ignore [reportIncompatibleVariableOverride]
        def call(self): ...

        field1: str = "tool_field"

    class ToolCallNoIDClass:
        pass

    class ToolCallNoID(BaseTool):
        tool_call: ClassVar[Any] = ToolCallNoIDClass()  # no id attribute

        def call(self): ...

        field1: str = "tool_field"

    class MockTool(Tool):
        def call(self): ...

    tool_with_id = MockTool(ToolWithID())
    tool_no_call = MockTool(ToolNoCall())
    tool_call_no_id = MockTool(ToolCallNoID())

    result_with_id = CallResponse.tool_message_params([(tool_with_id, "output1")])
    assert isinstance(result_with_id[0].content[0], ToolResultPart)
    assert result_with_id[0].content[0].id == "tool_call_with_id"

    result_no_call = CallResponse.tool_message_params([(tool_no_call, "output2")])
    assert isinstance(result_no_call[0].content[0], ToolResultPart)
    assert result_no_call[0].content[0].id is None

    result_no_id = CallResponse.tool_message_params([(tool_call_no_id, "output3")])
    assert isinstance(result_no_id[0].content[0], ToolResultPart)
    assert result_no_id[0].content[0].id is None


@pytest.mark.asyncio
async def test_tool_property_returns_first_tool(dummy_call_response_instance):
    first_tool = Tool(tool=DummyTool(MagicMock(spec=BaseTool)))  # pyright: ignore [reportAbstractUsage]

    with patch.object(
        type(dummy_call_response_instance._response),
        "common_tools",
        new_callable=PropertyMock,
    ) as mock_common_tools:
        mock_common_tools.return_value = [first_tool]

        assert dummy_call_response_instance.tool == first_tool, (
            "Expected the first Tool in _response.common_tools"
        )


@pytest.mark.asyncio
async def test_tool_returns_none_when_no_tools(dummy_call_response_instance):
    with patch.object(
        type(dummy_call_response_instance._response),
        "common_tools",
        new_callable=PropertyMock,
    ) as mock_common_tools:
        mock_common_tools.return_value = None

        assert dummy_call_response_instance.tool is None, (
            "Expected None when _response.common_tools is None"
        )



"""Tests the `_utils.get_prompt_template` module."""

import os

import pytest

from mirascope.core.base._utils._get_prompt_template import get_prompt_template
from mirascope.core.base.prompt import BasePrompt, prompt_template


def test_get_prompt_template() -> None:
    """Tests the `get_prompt_template` function."""

    class Prompt(BasePrompt): ...

    with pytest.raises(ValueError, match="No prompt template set!"):
        get_prompt_template(Prompt)

    assert get_prompt_template(prompt_template("prompt")(Prompt)) == "prompt"

    class PromptWithDocstring(BasePrompt):
        """docstring_prompt"""

    with pytest.raises(
        ValueError, match="You must explicitly enable docstring prompt templates."
    ):
        get_prompt_template(PromptWithDocstring)

    os.environ["MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE"] = "ENABLED"
    assert get_prompt_template(PromptWithDocstring) == "docstring_prompt"
    os.environ["MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE"] = "DISABLED"

    def fn() -> None: ...  # pragma: no cover

    with pytest.raises(ValueError, match="No prompt template set!"):
        get_prompt_template(fn)

    assert get_prompt_template(prompt_template("prompt")(fn)) == "prompt"

    def fn_with_docstring() -> None:
        """docstring_prompt"""

    with pytest.raises(
        ValueError, match="You must explicitly enable docstring prompt templates."
    ):
        get_prompt_template(fn_with_docstring)

    os.environ["MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE"] = "ENABLED"
    assert get_prompt_template(fn_with_docstring) == "docstring_prompt"
    os.environ["MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE"] = "DISABLED"



"""Protocols for reusable type hints."""

from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    Generator,
    Iterable,
)
from enum import Enum
from typing import (
    Any,
    Literal,
    NoReturn,
    ParamSpec,
    Protocol,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from ..call_kwargs import BaseCallKwargs
from ..call_response import BaseCallResponse
from ..call_response_chunk import BaseCallResponseChunk
from ..messages import Messages
from ..stream_config import StreamConfig
from ..tool import BaseTool
from ._base_type import BaseType

_BaseCallResponseT = TypeVar(
    "_BaseCallResponseT", covariant=True, bound=BaseCallResponse
)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", covariant=True, bound=BaseCallResponseChunk
)
_BaseStreamT = TypeVar("_BaseStreamT", covariant=True)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType | Enum)
_SameSyncAndAsyncClientT = TypeVar("_SameSyncAndAsyncClientT", contravariant=True)
_SyncBaseClientT = TypeVar("_SyncBaseClientT", contravariant=True)
_AsyncBaseClientT = TypeVar("_AsyncBaseClientT", contravariant=True)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", contravariant=True)
_AsyncBaseDynamicConfigT = TypeVar("_AsyncBaseDynamicConfigT", contravariant=True)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", contravariant=True)
_ResponseT = TypeVar("_ResponseT", covariant=True)
_AsyncResponseT = TypeVar("_AsyncResponseT", covariant=True)
_ResponseChunkT = TypeVar("_ResponseChunkT", covariant=True)
_AsyncResponseChunkT = TypeVar("_AsyncResponseChunkT", covariant=True)
_InvariantResponseChunkT = TypeVar("_InvariantResponseChunkT", contravariant=True)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_P = ParamSpec("_P")
_R = TypeVar("_R", contravariant=True)


class AsyncLLMFunctionDecorator(Protocol[_AsyncBaseDynamicConfigT, _AsyncResponseT]):
    @overload
    def __call__(
        self,
        fn: Callable[
            _P,
            Awaitable[_AsyncBaseDynamicConfigT]
            | Coroutine[Any, Any, _AsyncBaseDynamicConfigT],
        ],
    ) -> Callable[_P, Awaitable[_AsyncResponseT]]: ...

    @overload
    def __call__(
        self,
        fn: Callable[_P, Awaitable[Messages.Type] | Coroutine[Any, Any, Messages.Type]],
    ) -> Callable[_P, Awaitable[_AsyncResponseT]]: ...

    def __call__(
        self,
        fn: Callable[
            _P,
            Awaitable[_AsyncBaseDynamicConfigT]
            | Coroutine[Any, Any, _AsyncBaseDynamicConfigT],
        ]
        | Callable[_P, Awaitable[Messages.Type] | Coroutine[Any, Any, Messages.Type]],
    ) -> Callable[_P, Awaitable[_AsyncResponseT]]: ...  # pragma: no cover


class SyncLLMFunctionDecorator(Protocol[_BaseDynamicConfigT, _ResponseT]):
    @overload
    def __call__(
        self, fn: Callable[_P, _BaseDynamicConfigT]
    ) -> Callable[_P, _ResponseT]: ...

    @overload
    def __call__(self, fn: Callable[_P, Messages.Type]) -> Callable[_P, _ResponseT]: ...

    def __call__(
        self, fn: Callable[_P, _BaseDynamicConfigT] | Callable[_P, Messages.Type]
    ) -> Callable[_P, _ResponseT]: ...  # pragma: no cover


class LLMFunctionDecorator(
    Protocol[_BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _ResponseT, _AsyncResponseT]
):
    @overload
    def __call__(
        self, fn: Callable[_P, _BaseDynamicConfigT]
    ) -> Callable[_P, _ResponseT]: ...

    @overload
    def __call__(self, fn: Callable[_P, Messages.Type]) -> Callable[_P, _ResponseT]: ...

    @overload
    def __call__(
        self, fn: Callable[_P, Awaitable[_AsyncBaseDynamicConfigT]]
    ) -> Callable[_P, Awaitable[_AsyncResponseT]]: ...

    @overload
    def __call__(
        self, fn: Callable[_P, Awaitable[Messages.Type]]
    ) -> Callable[_P, Awaitable[_AsyncResponseT]]: ...

    def __call__(
        self,
        fn: Callable[_P, _BaseDynamicConfigT]
        | Callable[_P, Awaitable[_AsyncBaseDynamicConfigT]]
        | Callable[_P, Messages.Type]
        | Callable[_P, Awaitable[Messages.Type]],
    ) -> Callable[_P, _ResponseT | Awaitable[_AsyncResponseT]]: ...  # pragma: no cover


class AsyncCreateFn(Protocol[_ResponseT, _ResponseChunkT]):
    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        *,
        stream: Literal[False] = False,
        **kwargs: Any,  # noqa: ANN401
    ) -> Awaitable[_ResponseT]: ...

    @overload
    def __call__(
        self,
        *,
        stream: Literal[True] | StreamConfig = True,
        **kwargs: Any,  # noqa: ANN401
    ) -> Awaitable[AsyncGenerator[_ResponseChunkT, None]]: ...

    def __call__(
        self, *, stream: bool | StreamConfig = False, **kwargs: Any
    ) -> Awaitable[
        _ResponseT | AsyncGenerator[_ResponseChunkT, None]
    ]: ...  # pragma: no cover


class CreateFn(Protocol[_ResponseT, _ResponseChunkT]):
    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        *,
        stream: Literal[False] = False,
        **kwargs: Any,  # noqa: ANN401
    ) -> _ResponseT: ...

    @overload
    def __call__(
        self,
        *,
        stream: Literal[True] | StreamConfig = True,
        **kwargs: Any,  # noqa: ANN401
    ) -> Generator[_ResponseChunkT, None, None]: ...

    def __call__(
        self,
        *,
        stream: bool | StreamConfig = False,
        **kwargs: Any,  # noqa: F821
    ) -> _ResponseT | Generator[_ResponseChunkT, None, None]: ...


class SetupCall(
    Protocol[
        _SyncBaseClientT,
        _AsyncBaseClientT,
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ]
):
    @overload
    def __call__(
        self,
        *,
        model: str,
        client: _AsyncBaseClientT | None,
        fn: Callable[..., Awaitable[_AsyncBaseDynamicConfigT]],
        fn_args: dict[str, Any],
        dynamic_config: _AsyncBaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        response_model: type[BaseModel] | None,
        stream: bool | StreamConfig,
    ) -> tuple[
        AsyncCreateFn[_AsyncResponseT, _AsyncResponseChunkT],
        str | None,
        list[Any],
        list[type[_BaseToolT]] | None,
        BaseCallKwargs,
    ]: ...

    @overload
    def __call__(
        self,
        *,
        model: str,
        client: _SyncBaseClientT | None,
        fn: Callable[..., _BaseDynamicConfigT],
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        response_model: type[BaseModel] | None,
        stream: bool | StreamConfig,
    ) -> tuple[
        CreateFn[_ResponseT, _ResponseChunkT],
        str | None,
        list[Any],
        list[type[_BaseToolT]] | None,
        BaseCallKwargs,
    ]: ...

    def __call__(
        self,
        *,
        model: str,
        client: _SyncBaseClientT | _AsyncBaseClientT | None,
        fn: Callable[..., _BaseDynamicConfigT | Awaitable[_AsyncBaseDynamicConfigT]],
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT | _AsyncBaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        response_model: type[BaseModel] | None,
        stream: bool | StreamConfig,
    ) -> tuple[
        CreateFn[_ResponseT, _ResponseChunkT]
        | AsyncCreateFn[_AsyncResponseT, _AsyncResponseChunkT],
        str | None,
        list[Any],
        list[type[_BaseToolT]] | None,
        BaseCallKwargs,
    ]: ...  # pragma: no cover


class SameSyncAndAsyncClientSetupCall(
    Protocol[
        _SameSyncAndAsyncClientT,
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallParamsT,
        _ResponseT,
        _ResponseChunkT,
        _AsyncResponseT,
        _AsyncResponseChunkT,
        _BaseToolT,
    ]
):
    @overload
    def __call__(
        self,
        *,
        model: str,
        client: _SameSyncAndAsyncClientT | None,
        fn: Callable[..., Awaitable[_AsyncBaseDynamicConfigT]],
        fn_args: dict[str, Any],
        dynamic_config: _AsyncBaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        response_model: type[BaseModel] | None,
        stream: bool | StreamConfig,
    ) -> tuple[
        AsyncCreateFn[_AsyncResponseT, _AsyncResponseChunkT],
        str | None,
        list[Any],
        list[type[_BaseToolT]] | None,
        BaseCallKwargs,
    ]: ...

    @overload
    def __call__(
        self,
        *,
        model: str,
        client: _SameSyncAndAsyncClientT | None,
        fn: Callable[..., _BaseDynamicConfigT],
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        response_model: type[BaseModel] | None,
        stream: bool | StreamConfig,
    ) -> tuple[
        CreateFn[_ResponseT, _ResponseChunkT],
        str | None,
        list[Any],
        list[type[_BaseToolT]] | None,
        BaseCallKwargs,
    ]: ...

    def __call__(
        self,
        *,
        model: str,
        client: _SameSyncAndAsyncClientT | None,
        fn: Callable[..., _BaseDynamicConfigT | Awaitable[_AsyncBaseDynamicConfigT]],
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT | _AsyncBaseDynamicConfigT,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        call_params: _BaseCallParamsT,
        response_model: type[BaseModel] | None,
        stream: bool | StreamConfig,
    ) -> tuple[
        CreateFn[_ResponseT, _ResponseChunkT]
        | AsyncCreateFn[_AsyncResponseT, _AsyncResponseChunkT],
        str | None,
        list[Any],
        list[type[_BaseToolT]] | None,
        BaseCallKwargs,
    ]: ...  # pragma: no cover


class HandleStream(
    Protocol[_InvariantResponseChunkT, _BaseCallResponseChunkT, _BaseToolT]
):
    def __call__(
        self,
        stream: Generator[_InvariantResponseChunkT, None, None],
        tool_types: list[type[_BaseToolT]] | None,
        partial_tools: bool = False,
    ) -> Generator[
        tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None
    ]: ...  # pragma: no cover


class HandleStreamAsync(
    Protocol[_InvariantResponseChunkT, _BaseCallResponseChunkT, _BaseToolT]
):
    def __call__(
        self,
        stream: AsyncGenerator[_InvariantResponseChunkT, None],
        tool_types: list[type[_BaseToolT]] | None,
        partial_tools: bool = False,
    ) -> AsyncGenerator[
        tuple[_BaseCallResponseChunkT, _BaseToolT | None], None
    ]: ...  # pragma: no cover


class GetJsonOutput(Protocol[_R]):
    def __call__(self, response: _R, json_mode: bool) -> str: ...  # pragma: no cover


class CalculateCost(Protocol):
    def __call__(
        self,
        input_tokens: int | float | None,
        cached_tokens: int | float | None,
        output_tokens: int | float | None,
        model: str,
    ) -> float | None: ...  # pragma: no cover


class CallDecorator(
    Protocol[
        _BaseCallResponseT,
        _BaseCallResponseChunkT,
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallParamsT,
        _BaseStreamT,
        _SyncBaseClientT,
        _AsyncBaseClientT,
        _SameSyncAndAsyncClientT,
    ],
):
    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _BaseCallResponseT,
        _BaseCallResponseT,
    ]: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _BaseCallResponseT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _BaseCallResponseT]: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _ParsedOutputT, _ParsedOutputT
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseChunkT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _SyncBaseClientT
        | _AsyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _BaseStreamT, _BaseStreamT
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _BaseStreamT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _BaseStreamT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseChunkT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _SyncBaseClientT
        | _AsyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _SyncBaseClientT
        | _AsyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _ResponseModelT, _ResponseModelT
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _ResponseModelT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _ResponseModelT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT, _AsyncBaseDynamicConfigT, _ParsedOutputT, _ParsedOutputT
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[_AsyncBaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, _ParsedOutputT]: ...

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable],
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _ResponseModelT | _BaseCallResponseT,
        _ResponseModelT | _BaseCallResponseT,
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable],
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[
        _BaseDynamicConfigT,
        _ResponseModelT | _BaseCallResponseT,
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable],
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[
        _AsyncBaseDynamicConfigT,
        _ResponseModelT | _BaseCallResponseT,
    ]: ...

    ####

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable],
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        _ParsedOutputT | _BaseCallResponseT,
        _ParsedOutputT | _BaseCallResponseT,
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable],
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[
        _BaseDynamicConfigT,
        _ParsedOutputT | _BaseCallResponseT,
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable],
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_ResponseModelT], _ParsedOutputT],
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[
        _AsyncBaseDynamicConfigT,
        _ParsedOutputT | _BaseCallResponseT,
    ]: ...

    ####

    @overload
    def __call__(  # pyright: ignore[reportOverlappingOverload]
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> LLMFunctionDecorator[
        _BaseDynamicConfigT,
        _AsyncBaseDynamicConfigT,
        Iterable[_ResponseModelT],
        AsyncIterable[_ResponseModelT],
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _AsyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> AsyncLLMFunctionDecorator[
        _AsyncBaseDynamicConfigT, AsyncIterable[_ResponseModelT]
    ]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _SyncBaseClientT = ...,
        call_params: _BaseCallParamsT | None = None,
    ) -> SyncLLMFunctionDecorator[_BaseDynamicConfigT, Iterable[_ResponseModelT]]: ...

    @overload
    def __call__(
        self,
        model: str,
        *,
        stream: Literal[True] | StreamConfig,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT]
        | Callable[[_BaseCallResponseChunkT], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _AsyncBaseClientT
        | _SyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> NoReturn: ...

    def __call__(
        self,
        model: str,
        *,
        stream: bool | StreamConfig = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT] | None = None,
        output_parser: Callable[[_BaseCallResponseT], _ParsedOutputT]
        | Callable[[_BaseCallResponseChunkT], _ParsedOutputT]
        | Callable[[_ResponseModelT], _ParsedOutputT]
        | None = None,
        json_mode: bool = False,
        client: _SameSyncAndAsyncClientT
        | _AsyncBaseClientT
        | _SyncBaseClientT
        | None = None,
        call_params: _BaseCallParamsT | None = None,
    ) -> (
        AsyncLLMFunctionDecorator[
            _AsyncBaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | (_ResponseModelT | _BaseCallResponseT)
            | AsyncIterable[_ResponseModelT],
        ]
        | SyncLLMFunctionDecorator[
            _BaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | (_ResponseModelT | _BaseCallResponseT)
            | (_ParsedOutputT | _BaseCallResponseT)
            | Iterable[_ResponseModelT],
        ]
        | LLMFunctionDecorator[
            _BaseDynamicConfigT,
            _AsyncBaseDynamicConfigT,
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | (_ResponseModelT | _BaseCallResponseT)
            | (_ParsedOutputT | _BaseCallResponseT)
            | Iterable[_ResponseModelT],
            _BaseCallResponseT
            | _ParsedOutputT
            | _BaseStreamT
            | _ResponseModelT
            | (_ResponseModelT | _BaseCallResponseT)
            | (_ParsedOutputT | _BaseCallResponseT)
            | AsyncIterable[_ResponseModelT],
        ]
    ): ...



from io import BytesIO

import pytest
from PIL import Image

from mirascope.core.base._utils._pil_image_to_bytes import pil_image_to_bytes


@pytest.fixture
def mock_jpeg_bytes() -> bytes:
    return b"\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01\x00\x00\x01\x00\x01\x00\x00\xff\xdb\x00C\x00\x08\x06\x06\x07\x06\x05\x08\x07\x07\x07\t\t\x08\n\x0c\x14\r\x0c\x0b\x0b\x0c\x19\x12\x13\x0f\x14\x1d\x1a\x1f\x1e\x1d\x1a\x1c\x1c $.' \",#\x1c\x1c(7),01444\x1f'9=82<.342\xff\xc0\x00\x0b\x08\x00\x01\x00\x01\x01\x01\x11\x00\xff\xc4\x00\x1f\x00\x00\x01\x05\x01\x01\x01\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\xff\xc4\x00\xb5\x10\x00\x02\x01\x03\x03\x02\x04\x03\x05\x05\x04\x04\x00\x00\x01}\x01\x02\x03\x00\x04\x11\x05\x12!1A\x06\x13Qa\x07\"q\x142\x81\x91\xa1\x08#B\xb1\xc1\x15R\xd1\xf0$3br\x82\t\n\x16\x17\x18\x19\x1a%&'()*456789:CDEFGHIJSTUVWXYZcdefghijstuvwxyz\x83\x84\x85\x86\x87\x88\x89\x8a\x92\x93\x94\x95\x96\x97\x98\x99\x9a\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xff\xda\x00\x08\x01\x01\x00\x00?\x00\xf9\xfe\xbf\xff\xd9"


@pytest.fixture
def mock_pil_image(mock_jpeg_bytes: bytes) -> Image.Image:
    return Image.open(BytesIO(mock_jpeg_bytes))


def test_pil_image_to_bytes_success_jpg(
    mock_jpeg_bytes: bytes, mock_pil_image: Image.Image
):
    """Tests pil_image_to_bytes with a JPEG image (mocking format)."""
    image_bytes = pil_image_to_bytes(mock_pil_image)
    assert isinstance(image_bytes, bytes)
    assert image_bytes == mock_jpeg_bytes


def test_pil_image_to_bytes_no_format(mock_pil_image: Image.Image):
    """
    Tests pil_image_to_bytes with an image with no format attribute.
    """
    mock_pil_image.format = None
    with pytest.raises(ValueError, match="Error converting image to bytes:"):
        pil_image_to_bytes(mock_pil_image)



"""This module contains the base class for message parameters."""

from collections.abc import Sequence
from typing import Literal

from pydantic import BaseModel


class TextPart(BaseModel):
    """A content part for text.

    Attributes:
        type: Always "text"
        text: The text content
    """

    type: Literal["text"]
    text: str


class CacheControlPart(BaseModel):
    """A part for marking cache control.

    This part is currently only available with Anthropic. For more details, see:
    https://docs.anthropic.com/en/docs/build-with-claude/prompt-caching

    Attributes:
        type: Always "cache_control"
        cache_type: Currently only "ephemeral" is available.
    """

    type: Literal["cache_control"]
    cache_type: str


class ImagePart(BaseModel):
    """A content part for images.

    Attributes:
        type: Always "image"
        media_type: The media type (e.g. image/jpeg)
        image: The raw image bytes
        detail: (Optional) The detail to use for the image (supported by OpenAI)
    """

    type: Literal["image"]
    media_type: str
    image: bytes
    detail: str | None


class ImageURLPart(BaseModel):
    """A content part for images with a URL or base64 encoded image data.

    Attributes:
        type: Always "image_url"
        url: The URL to the image
        detail: (Optional) The detail to use for the image (supported by OpenAI)
    """

    type: Literal["image_url"]
    url: str
    detail: str | None


class AudioPart(BaseModel):
    """A content part for audio.

    Attributes:
        type: Always "audio"
        media_type: The media type (e.g. audio/wav)
        audio: The raw audio bytes or base64 encoded audio data
    """

    type: Literal["audio"]
    media_type: str
    audio: bytes | str


class AudioURLPart(BaseModel):
    """A content part for audio with a URL or base64 encoded audio data.

    Attributes:
        type: Always "audio_url"
        url: The URL to the audio
    """

    type: Literal["audio_url"]
    url: str


class DocumentPart(BaseModel):
    """A content part for pdf.

    Attributes:
        type: Always "document"
        media_type: The media type (e.g. application/pdf)
        document: document data
    """

    type: Literal["document"]
    media_type: str
    document: bytes


class DocumentURLPart(BaseModel):
    """A content part for documents with a URL.

    Attributes:
        type: Always "document_url"
        url: The URL to the document
    """

    type: Literal["document_url"]
    url: str


class ToolCallPart(BaseModel):
    """A content part for tool.

    Attributes:
        type: Always "tool"
        name: The name of the tool
        id: The id of the tool
    """

    type: Literal["tool_call"]
    name: str
    args: dict | None = None
    id: str | None = None


class ToolResultPart(BaseModel):
    """A content part for tool.

    Attributes:
        type: Always "tool"
        name: The name of the tool
        id: The id of the tool
    """

    type: Literal["tool_result"]
    name: str = ""
    content: str
    id: str | None = None
    is_error: bool = False


class BaseMessageParam(BaseModel):
    """A base class for message parameters.

    usage docs: learn/prompts.md#prompt-templates-messages

    Attributes:
        role: The role of the message (e.g. "system", "user", "assistant", "tool")
        content: The content of the message
        tool_name: The name of the tool, if any
    """

    role: str
    content: (
        str
        | Sequence[
            TextPart
            | ImagePart
            | ImageURLPart
            | AudioPart
            | AudioURLPart
            | CacheControlPart
            | DocumentPart
            | DocumentURLPart
            | ToolCallPart
            | ToolResultPart
        ]
    )



from google.genai.types import GenerateContentConfig

from mirascope.core.base.call_params import CommonCallParams
from mirascope.core.google._utils._convert_common_call_params import (
    convert_common_call_params,
)


def test_google_conversion_full():
    """Test full parameter conversion for Google."""
    params: CommonCallParams = {
        "temperature": 0.7,
        "max_tokens": 100,
        "top_p": 0.9,
        "stop": ["STOP", "END"],
    }
    result = convert_common_call_params(params)

    config = result.get("config")
    assert isinstance(config, GenerateContentConfig)
    assert config.temperature == 0.7
    assert config.max_output_tokens == 100
    assert config.top_p == 0.9
    assert config.stop_sequences == ["STOP", "END"]


def test_google_conversion_single_stop():
    """Test single stop sequence conversion for Google."""
    params: CommonCallParams = {
        "temperature": 0.7,
        "stop": "STOP",
    }
    result = convert_common_call_params(params)

    config = result.get("config")
    assert isinstance(config, GenerateContentConfig)
    assert config.temperature == 0.7
    assert config.stop_sequences == ["STOP"]


def test_google_conversion_empty():
    """Test empty parameters conversion for Google."""
    result = convert_common_call_params({})
    assert result == {}


def test_google_conversion_full_with_invalid_key():
    """Test full parameter conversion for Google."""
    params: CommonCallParams = {
        "temperature": 0.7,
        "max_tokens": 100,
        "top_p": 0.9,
        "stop": ["STOP", "END"],
        "invalid": "invalid",  # pyright: ignore [reportAssignmentType]
    }
    result = convert_common_call_params(params)

    config = result.get("config")
    assert isinstance(config, GenerateContentConfig)
    assert config.temperature == 0.7
    assert config.max_output_tokens == 100
    assert config.top_p == 0.9
    assert config.stop_sequences == ["STOP", "END"]



"""Tests for the `toolkit` module."""

from typing import ClassVar, Literal
from unittest import mock

import pytest

from mirascope.core.base import BaseToolKit, toolkit_tool


@pytest.fixture
def mock_namespaces():
    mock_namespaces = set()
    with mock.patch("mirascope.core.base.toolkit._namespaces", mock_namespaces):
        yield mock_namespaces


@pytest.mark.parametrize(
    "namespace, expected_name",
    [
        (None, "format_book"),
        ("book_tools", "book_tools_format_book"),
    ],
)
def test_toolkit(mock_namespaces, namespace: str | None, expected_name: str) -> None:
    """Tests the `BaseToolKit` class and the `toolkit_tool` decorator."""

    class BookRecommendationToolKit(BaseToolKit):
        """A toolkit for recommending books."""

        __namespace__: ClassVar[str | None] = namespace
        reading_level: Literal["beginner", "advanced"]

        @toolkit_tool
        def format_book(self, title: str, author: str) -> str:
            """Returns formatted title and author.

            Reading level: {self.reading_level}
            """
            return f"{title} by {author}"

    toolkit = BookRecommendationToolKit(reading_level="beginner")
    tools = toolkit.create_tools()
    assert len(tools) == 1
    tool = tools[0]
    assert tool._name() == expected_name
    assert (
        tool._description()
        == "Returns formatted title and author.\n\nReading level: beginner"
    )
    assert (
        tool(title="The Name of the Wind", author="Rothfuss, Patrick").call()  # type: ignore
        == "The Name of the Wind by Rothfuss, Patrick"
    )


def test_toolkit_multiple_method(mock_namespaces) -> None:
    """Tests that multiple `toolkit_tool` methods are created correctly."""

    def dummy_decorator(func):
        return func

    class BookRecommendationToolKit(BaseToolKit):
        """A toolkit for recommending books."""

        __namespace__: ClassVar[str | None] = "book_tools"
        reading_level: Literal["beginner", "advanced"]
        language: Literal["english", "spanish", "french"]

        @toolkit_tool
        def format_book(self, title: str, author: str) -> str:
            """Returns formatted title and author.

            Reading level: {self.reading_level}
            """
            return f"{title} by {author}"

        @toolkit_tool
        def format_world_book(self, title: str, author: str, genre: str) -> str:
            """Returns formatted title, author, and genre.

            Reading level: {self.reading_level}
            language: {self.language}
            """
            return f"{title} by {author} ({genre})"

        @dummy_decorator
        def dummy_method(self) -> str:
            """dummy method"""
            return "dummy"  # pragma: no cover

    toolkit = BookRecommendationToolKit(reading_level="beginner", language="spanish")
    tools = toolkit.create_tools()
    assert len(tools) == 2

    assert tools[0]._name() == "book_tools_format_book"
    assert (
        tools[0]._description()
        == "Returns formatted title and author.\n\nReading level: beginner"
    )
    assert (
        tools[0](title="The Name of the Wind", author="Rothfuss, Patrick").call()  # type: ignore
        == "The Name of the Wind by Rothfuss, Patrick"
    )
    assert tools[1]._name() == "book_tools_format_world_book"
    assert (
        tools[1]._description()
        == "Returns formatted title, author, and genre.\n\nReading level: beginner\n"
        "language: spanish"
    )
    assert (
        tools[1](
            title="The Name of the Wind",  # type: ignore
            author="Rothfuss, Patrick",  # type: ignore
            genre="fantasy",  # type: ignore
        ).call()
        == "The Name of the Wind by Rothfuss, Patrick (fantasy)"
    )


def test_toolkit_tool_method_not_found() -> None:
    """Tests that a ValueError is raised when there's no `toolkit_tool` method."""

    def dummy_decorator(func):
        return func

    with pytest.raises(ValueError, match="No toolkit_tool method found"):

        class BookRecommendationToolKit(BaseToolKit):
            """A toolkit for recommending books."""

            __namespace__: ClassVar[str | None] = "book_tools"
            reading_level: Literal["beginner", "advanced"]
            language: Literal["english", "spanish", "french"]

            def format_book(self, title: str, author: str) -> str:
                """Returns the title and author of a book nicely formatted.

                Reading level: {self.reading_level}
                """
                return f"{title} by {author}"  # pragma: no cover

            @dummy_decorator
            def dummy_method(self) -> str:
                """dummy method"""
                return "dummy"  # pragma: no cover


def test_toolkit_tool_method_with_non_self_var(mock_namespaces) -> None:
    """Tests that non-self template variables are preserved through tool creation."""

    class BookRecommendationToolKit(BaseToolKit):
        """A toolkit for recommending books."""

        __namespace__: ClassVar[str | None] = "book_tools"
        reading_level: Literal["beginner", "advanced"]

        @toolkit_tool
        def format_book(self, title: str, author: str) -> str:
            """Reading Level: {self.reading_level}, Non-Self Template Var: {foo}"""
            return f"{title} by {author}"  # pragma: no cover

    tool = BookRecommendationToolKit(reading_level="beginner").create_tools()[0]
    assert (
        tool._description() == "Reading Level: beginner, Non-Self Template Var: {foo}"
    )


def test_toolkit_tool_method_has_no_exists_var(mock_namespaces) -> None:
    """Tests that a ValueError is raised when a template variable doesn't exist."""

    with pytest.raises(
        ValueError,
        match="The toolkit_tool method template variable self.not_exists is not found "
        "in the class",
    ):

        class BookRecommendationToolKit(BaseToolKit):
            """A toolkit for recommending books."""

            __namespace__: ClassVar[str | None] = "book_tools"
            reading_level: Literal["beginner", "advanced"]
            language: Literal["english", "spanish", "french"]

            @toolkit_tool
            def format_book(self, title: str, author: str) -> str:
                """Returns the title and author of a book nicely formatted.

                Reading level: {self.not_exists}
                """
                return f"{title} by {author}"  # pragma: no cover


def test_toolkit_namespace_already_used(mock_namespaces) -> None:
    """Tests that a ValueError is raised when the namespace is already used."""

    mock_namespaces.add("book_tools")
    with pytest.raises(ValueError, match="The namespace book_tools is already used"):

        class BookRecommendationToolKit(BaseToolKit):
            """A toolkit for recommending books."""

            __namespace__: ClassVar[str | None] = "book_tools"


def test_toolkit_tool_method_has_no_docstring(mock_namespaces) -> None:
    """Tests that a ValueError is raised when `toolkit_tool` method has no docstring."""

    with pytest.raises(
        ValueError, match="The toolkit_tool method must have a docstring"
    ):

        class BookRecommendationToolKit(BaseToolKit):
            @toolkit_tool
            def format_book(self, title: str, author: str) -> str:
                return f"{title} by {author}"  # pragma: no cover



"""This module defines the base class for tools used in LLM calls.

usage-docs: learn/tools.md
"""

from __future__ import annotations

import inspect
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, ClassVar, TypeVar

import jiter
from pydantic import BaseModel, ConfigDict
from pydantic.json_schema import (
    DEFAULT_REF_TEMPLATE,
    GenerateJsonSchema,
    JsonSchemaMode,
    JsonSchemaValue,
    SkipJsonSchema,
)
from pydantic_core.core_schema import CoreSchema
from typing_extensions import TypedDict

from . import _utils

_BaseToolT = TypeVar("_BaseToolT", bound=BaseModel)
_ToolSchemaT = TypeVar("_ToolSchemaT")


class ToolConfig(TypedDict, total=False):
    """A base class for tool configurations."""


class GenerateJsonSchemaNoTitles(GenerateJsonSchema):
    _openai_strict: ClassVar[bool] = False

    def _remove_title(self, key: str | None, obj: Any) -> Any:  # noqa: ANN401
        if isinstance(obj, dict):
            if self._openai_strict and "type" in obj and obj["type"] == "object":
                obj["additionalProperties"] = False
            if "type" in obj or "$ref" in obj or "properties" in obj:
                title = obj.pop("title", None)
                if key and title and key.lower() != title.lower():
                    obj["title"] = title

            for key, value in list(obj.items()):
                obj[key] = self._remove_title(key, value)
        elif isinstance(obj, list):
            return [self._remove_title(None, item) for item in obj]

        return obj

    def generate(
        self, schema: CoreSchema, mode: JsonSchemaMode = "validation"
    ) -> JsonSchemaValue:
        json_schema = super().generate(schema, mode=mode)
        json_schema.pop("title", None)
        json_schema.pop("description", None)
        json_schema = self._remove_title(None, json_schema)
        return json_schema


class BaseTool(BaseModel, ABC):
    '''A class for defining tools for LLM calls.

    Example:

    ```python
    from mirascope.core import BaseTool
    from pydantic import Field


    class FormatBook(BaseTool):
        """Returns a nicely formatted book recommendation."""

        title: str = Field(..., description="The title of the book.")
        author: str = Field(..., description="The author of the book.")

        def call(self) -> str:
            return f"{self.title} by {self.author}"
    ```
    '''

    __provider__: ClassVar[str] = "NONE"
    __tool_config_type__: ClassVar[type[ToolConfig]] = ToolConfig
    __custom_name__: ClassVar[str] = ""
    tool_config: ClassVar[ToolConfig] = ToolConfig()
    model_config = ConfigDict(arbitrary_types_allowed=True)
    delta: SkipJsonSchema[str | None] = None

    @classmethod
    def _dict_from_json(cls, json: str, allow_partial: bool = False) -> dict[str, Any]:
        """Returns a dictionary from a JSON string."""
        if not json:
            return {}
        return jiter.from_json(
            json.encode(), partial_mode="trailing-strings" if allow_partial else "off"
        )

    @classmethod
    def _name(cls) -> str:
        """Returns the name of the tool."""
        return cls.__custom_name__ if cls.__custom_name__ else cls.__name__

    @classmethod
    def _description(cls) -> str:
        """Returns the description of the tool."""
        return (
            inspect.cleandoc(cls.__doc__)
            if cls.__doc__
            else _utils.DEFAULT_TOOL_DOCSTRING
        )

    @property
    def args(self) -> dict[str, Any]:
        """The arguments of the tool as a dictionary."""
        return {
            field: getattr(self, field)
            for field in self.model_fields
            if field not in {"tool_call", "delta"}
        }

    @property
    def id(self) -> str | None:
        """The id of the tool."""
        if tool_call := getattr(self, "tool_call", None):
            # Expect tool_call has an id attribute.
            # If not, we should override this method on the provider tool
            return getattr(tool_call, "id", None)
        return None

    @abstractmethod
    def call(self, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        """The method to call the tool."""
        ...

    @classmethod
    def type_from_fn(cls: type[_BaseToolT], fn: Callable) -> type[_BaseToolT]:
        """Returns this tool type converted from a function.

        Args:
            fn: The function to convert into this tool type.
        """
        return _utils.convert_function_to_base_tool(fn, cls)

    @classmethod
    def type_from_base_model_type(
        cls: type[_BaseToolT], tool_type: type[BaseModel]
    ) -> type[_BaseToolT]:
        """Returns this tool type converted from a given base tool type.

        Args:
            tool_type: The tool type to convert into this tool type. This can be a
                custom `BaseTool` or `BaseModel` definition.
        """
        return _utils.convert_base_model_to_base_tool(tool_type, cls)

    @classmethod
    def type_from_base_type(
        cls: type[_BaseToolT], base_type: type[_utils.BaseType]
    ) -> type[_BaseToolT]:
        """Returns this tool type converted from a base type.

        Args:
            base_type: The base type (e.g. `int`) to convert into this tool type.
        """
        return _utils.convert_base_type_to_base_tool(base_type, cls)

    @classmethod
    def tool_schema(cls) -> Any:  # noqa: ANN401
        raise RuntimeError(
            f"{cls.__name__}.tool_schema() is not implemented. "
            "This method should be implemented in provider-specific tool classes."
        )

    @classmethod
    def model_json_schema(
        cls,
        by_alias: bool = True,
        ref_template: str = DEFAULT_REF_TEMPLATE,
        schema_generator: type[GenerateJsonSchema] = GenerateJsonSchemaNoTitles,
        mode: JsonSchemaMode = "validation",
    ) -> dict[str, Any]:
        """Returns the generated JSON schema for the class."""
        cls.warn_for_unsupported_configurations()
        return super().model_json_schema(
            by_alias=by_alias,
            ref_template=ref_template,
            schema_generator=schema_generator,
            mode=mode,
        )

    @classmethod
    def warn_for_unsupported_configurations(cls) -> None:
        """Warns when a specific provider does not support provided config options."""
        unsupported_tool_keys = _utils.get_unsupported_tool_config_keys(
            cls.tool_config, cls.__tool_config_type__
        )
        if unsupported_tool_keys:
            warnings.warn(
                f"{cls.__provider__} does not support the following tool "
                f"configurations, so they will be ignored: {unsupported_tool_keys}",
                UserWarning,
            )

        if "strict" in cls.model_config and cls.__provider__ not in ["openai", "azure"]:
            warnings.warn(
                f"{cls.__provider__} does not support strict structured outputs, but "
                "you have configured `strict=True` in your `ResponseModelConfigDict`. "
                "Ignoring `strict` as this feature is only supported by OpenAI.",
                UserWarning,
            )



import json

import pytest
from openai.types.chat import ChatCompletionAssistantMessageParam

from mirascope.core import BaseMessageParam
from mirascope.core.base import (
    AudioPart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
)
from mirascope.core.openai._utils._message_param_converter import (
    OpenAIMessageParamConverter,
)


def test_convert_message_param_str_content():
    """
    If content is a string, produce role="assistant" with that string as content.
    """
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant", content="Hello, world!"
    )
    results = OpenAIMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert result.content == "Hello, world!"


def test_convert_message_param_list_content_with_text():
    """
    If content is a list of text blocks, produce multiple TextParts.
    """
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant",
        content=[{"text": "Hello", "type": "text"}, {"text": "World", "type": "text"}],
    )
    results = OpenAIMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert len(result.content) == 2
    assert isinstance(result.content[0], TextPart)
    assert result.content[0].text == "Hello"


def test_convert_message_param_list_content_with_refusal():
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant",
        content=[{"refusal": "I cannot do that", "type": "refusal"}],
    )
    with pytest.raises(ValueError, match="I cannot do that"):
        OpenAIMessageParamConverter.from_provider([message_param])


def test_convert_message_param_tool_calls():
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant",
        content=[{"type": "text", "text": "Some content"}],
        tool_calls=[
            {
                "type": "function",
                "function": {
                    "name": "test_tool",
                    "arguments": json.dumps({"key": "value"}),
                },
                "id": "tool_call_1",
            }
        ],
    )
    results = OpenAIMessageParamConverter.from_provider([message_param])
    result = results[0]
    assert len(result.content) == 2
    assert isinstance(result.content[1], ToolCallPart)


def test_convert_message_param_no_content_no_tool_calls():
    message_param = ChatCompletionAssistantMessageParam(role="assistant", content=None)
    results = OpenAIMessageParamConverter.from_provider([message_param])
    assert not results


def test_to_provider():
    message_param = BaseMessageParam(
        role="assistant",
        content=[
            TextPart(type="text", text="Hello, world!"),
            ToolCallPart(type="tool_call", name="test_tool", args={"key": "value"}),
            ToolCallPart(type="tool_call", name="test_tool_2", args={"key": "value"}),
        ],
    )
    results = OpenAIMessageParamConverter.to_provider([message_param])
    assert results == [
        {
            "content": [{"text": "Hello, world!", "type": "text"}],
            "role": "assistant",
            "tool_calls": [
                {
                    "function": {
                        "arguments": '{"key": "value"}',
                        "name": "test_tool",
                    },
                    "id": None,
                    "type": "function",
                },
                {
                    "function": {
                        "arguments": '{"key": "value"}',
                        "name": "test_tool_2",
                    },
                    "id": None,
                    "type": "function",
                },
            ],
        }
    ]


def test_convert_message_param_tool_result():
    message_param = ChatCompletionAssistantMessageParam(  # pyright: ignore [reportCallIssue]
        role="tool",
        name="test_tool",
        content="result",
        tool_call_id="tool_call_1",
    )
    results = OpenAIMessageParamConverter.from_provider([message_param])
    result = results[0]
    assert len(result.content) == 1
    assert result.content[0].name == "test_tool"  # pyright: ignore [reportAttributeAccessIssue]
    assert result.content[0].content == "result"  # pyright: ignore [reportAttributeAccessIssue]
    assert result.content[0].id == "tool_call_1"  # pyright: ignore [reportAttributeAccessIssue]
    assert not result.content[0].is_error  # pyright: ignore [reportAttributeAccessIssue]
    assert result.role == "tool"


def test_convert_message_param_list_content_with_image_url():
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant",
        content=[
            {  # pyright: ignore [reportArgumentType]
                "type": "image_url",
                "image_url": {"url": "https://example.com/image", "detail": "custom"},
            }
        ],
    )
    results = OpenAIMessageParamConverter.from_provider([message_param])
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    part = result.content[0]
    assert isinstance(part, ImageURLPart)
    assert part.url == "https://example.com/image"
    assert part.detail == "custom"


def test_convert_message_param_list_content_with_input_audio():
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant",
        content=[
            {  # pyright: ignore [reportArgumentType]
                "type": "input_audio",
                "input_audio": {"format": "mp3", "data": "YXVkaW8="},
            }
        ],
    )
    results = OpenAIMessageParamConverter.from_provider([message_param])
    assert len(results) == 1
    result = results[0]
    assert isinstance(result.content, list)
    assert len(result.content) == 1
    part = result.content[0]
    assert isinstance(part, AudioPart)
    assert part.media_type == "audio/mp3"
    assert part.audio == "YXVkaW8="



"""Tests the `vertex._utils.convert_message_params` function."""

import io
from unittest.mock import MagicMock, patch

import pytest
from vertexai.generative_models import Content, Part

from mirascope.core.base import (
    AudioPart,
    AudioURLPart,
    BaseMessageParam,
    CacheControlPart,
    ImagePart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.vertex._utils._convert_message_params import convert_message_params


@patch("PIL.Image.open", new_callable=MagicMock)
def test_convert_message_params(mock_image_open: MagicMock) -> None:
    """Tests the `convert_message_params` function."""
    mock_image_open.return_value.format = "jpeg"
    message_params: list[BaseMessageParam | Content] = [
        BaseMessageParam(role="system", content="You are a helpful assistant."),
        BaseMessageParam(role="user", content="Hello"),
        Content(role="user", parts=[Part.from_text("Hello"), Part.from_text("Hello")]),
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="test"),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail=None
                ),
                AudioPart(type="audio", media_type="audio/wav", audio=b"audio"),
                ToolResultPart(
                    name="tool_name", id="tool_id", content="result", type="tool_result"
                ),
                TextPart(type="text", text="test"),
                ToolCallPart(type="tool_call", name="tool_name", args={"arg": "val"}),
                TextPart(type="text", text="test"),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert [p.to_dict() for p in converted_message_params] == [
        {"parts": [{"text": "You are a helpful assistant."}], "role": "system"},
        {"parts": [{"text": "Hello"}], "role": "user"},
        {"parts": [{"text": "Hello"}, {"text": "Hello"}], "role": "user"},
        {
            "parts": [
                {"text": "test"},
                {"inline_data": {"data": "aW1hZ2U=", "mime_type": "image/jpeg"}},
                {"inline_data": {"data": "YXVkaW8=", "mime_type": "audio/wav"}},
            ],
            "role": "user",
        },
        {
            "parts": [
                {
                    "function_response": {
                        "name": "tool_name",
                        "response": {"content": {"result": "result"}},
                    }
                }
            ],
            "role": "user",
        },
        {"parts": [{"text": "test"}], "role": "user"},
        {
            "parts": [{"function_call": {"args": {"arg": "val"}, "name": "tool_name"}}],
            "role": "user",
        },
        {"parts": [{"text": "test"}], "role": "user"},
    ]
    mock_image_open.assert_called_once()
    bytes_io: io.BytesIO = mock_image_open.call_args.args[0]
    assert bytes_io.getvalue() == b"image"

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. Vertex currently only supports "
        "JPEG, PNG, WebP, HEIC, and HEIF images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail=None,
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Failed to load or encode data from http://example.com/image",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImageURLPart(
                            type="image_url",
                            url="http://example.com/image",
                            detail=None,
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/unknown. "
        "Vertex currently only supports WAV, MP3, AIFF, AAC, OGG, "
        "and FLAC audio file types.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(
                            type="audio", media_type="audio/unknown", audio=b"audio"
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Vertex currently only supports text, image, and audio parts. Part provided: cache_control",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        CacheControlPart(type="cache_control", cache_type="ephemeral")
                    ],
                )
            ]
        )


@patch(
    "mirascope.core.vertex._utils._convert_message_params._load_media",
    return_value=b"imagedata",
)
@patch("PIL.Image.open", new_callable=MagicMock)
@patch("mirascope.core.vertex._utils._convert_message_params.Part.from_uri")
def test_image_url_conversion(
    mock_from_uri, mock_image_open: MagicMock, mock_load_media
) -> None:
    """Tests conversion of an image_url part."""
    # Patch PIL.Image.MIME so that the lookup for "JPEG" succeeds.
    with patch.dict("PIL.Image.MIME", {"JPEG": "image/jpeg"}):
        dummy_image = MagicMock()
        dummy_image.format = "JPEG"
        mock_image_open.return_value = dummy_image

        # Define a dummy Part to be returned by Part.from_uri.
        class DummyPart:
            def __init__(self, data):
                # Set _raw_part to mimic a valid Part with file_data.
                self._raw_part = {
                    "file_data": {
                        "file_uri": data["file_uri"],
                        "mime_type": data["mime_type"],
                    }
                }

            def to_dict(self):
                return self._raw_part

        expected_dummy = DummyPart(
            {"file_uri": "http://example.com/image", "mime_type": "image/jpeg"}
        )
        mock_from_uri.return_value = expected_dummy

        # Create an instance of ImageURLPart.
        dummy_part = ImageURLPart(
            type="image_url", url="http://example.com/image", detail="auto"
        )
        message = BaseMessageParam(role="user", content=[dummy_part])
        result = convert_message_params([message])
        result_dicts = [p.to_dict() for p in result]
        expected = {"role": "user", "parts": [expected_dummy.to_dict()]}
        assert result_dicts == [expected]
        mock_load_media.assert_called_once_with("http://example.com/image")
        mock_from_uri.assert_called_once_with(
            "http://example.com/image", mime_type="image/jpeg"
        )


@patch(
    "mirascope.core.vertex._utils._convert_message_params._load_media",
    return_value=b"imagedata",
)
@patch("PIL.Image.open", new_callable=MagicMock)
@patch("mirascope.core.vertex._utils._convert_message_params.Part.from_uri")
def test_image_url_conversion_with_format_invalid(
    mock_from_uri, mock_image_open: MagicMock, mock_load_media
) -> None:
    """Tests conversion of an image_url part with format invalid."""
    with patch.dict("PIL.Image.MIME", {"JPEG": "image/invalid"}):
        dummy_image = MagicMock()
        dummy_image.format = "JPEG"
        mock_image_open.return_value = dummy_image

        # Define a dummy Part to be returned by Part.from_uri.
        class DummyPart:
            def __init__(self, data):
                # Set _raw_part to mimic a valid Part with file_data.
                self._raw_part = {
                    "file_data": {
                        "file_uri": data["file_uri"],
                        "mime_type": data["mime_type"],
                    }
                }

            def to_dict(self): ...

        expected_dummy = DummyPart(
            {"file_uri": "http://example.com/image", "mime_type": "image/jpeg"}
        )
        mock_from_uri.return_value = expected_dummy

        # Create an instance of ImageURLPart.
        dummy_part = ImageURLPart(
            type="image_url", url="http://example.com/image", detail="auto"
        )
        message = BaseMessageParam(role="user", content=[dummy_part])
        with pytest.raises(
            ValueError,
            match="Unsupported image media type: image/invalid. Gemini currently only supports JPEG, PNG, WebP, HEIC, and HEIF images.",
        ):
            convert_message_params([message])


@patch(
    "mirascope.core.vertex._utils._convert_message_params._load_media",
    return_value=b"audiodata",
)
@patch(
    "mirascope.core.vertex._utils._convert_message_params.get_audio_type",
    return_value="mp3",
)
@patch("mirascope.core.vertex._utils._convert_message_params.Part.from_uri")
def test_audio_url_conversion_valid(
    mock_from_uri, mock_get_audio_type, mock_load_media
) -> None:
    """Tests conversion of a valid audio_url part."""

    class DummyPart:
        def __init__(self, data):
            # Mimic a valid Part with file_data.
            self._raw_part = {
                "file_data": {
                    "file_uri": data["file_uri"],
                    "mime_type": data["mime_type"],
                }
            }

        def to_dict(self):
            return self._raw_part

    expected_dummy = DummyPart(
        {"file_uri": "http://example.com/audio", "mime_type": "audio/mp3"}
    )
    mock_from_uri.return_value = expected_dummy

    # Create an instance of AudioURLPart.
    dummy_part = AudioURLPart(type="audio_url", url="http://example.com/audio")
    message = BaseMessageParam(role="user", content=[dummy_part])
    result = convert_message_params([message])
    result_dicts = [p.to_dict() for p in result]
    expected = {"role": "user", "parts": [expected_dummy.to_dict()]}
    assert result_dicts == [expected]
    mock_load_media.assert_called_once_with("http://example.com/audio")
    mock_get_audio_type.assert_called_once_with(b"audiodata")
    mock_from_uri.assert_called_once_with(
        "http://example.com/audio", mime_type="audio/mp3"
    )


@patch(
    "mirascope.core.vertex._utils._convert_message_params._load_media",
    return_value=b"audiodata",
)
@patch(
    "mirascope.core.vertex._utils._convert_message_params.get_audio_type",
    return_value="invalid",
)
def test_audio_url_conversion_invalid(mock_get_audio_type, mock_load_media) -> None:
    """Tests conversion of an audio_url part with an unsupported audio type."""
    dummy_part = AudioURLPart(type="audio_url", url="http://example.com/audio")
    message = BaseMessageParam(role="user", content=[dummy_part])
    with pytest.raises(
        ValueError,
        match=r"Unsupported audio media type: audio/invalid\. Gemini currently only supports WAV, MP3, AIFF, AAC, OGG, and FLAC audio file types\.",
    ):
        convert_message_params([message])
    mock_load_media.assert_called_once_with("http://example.com/audio")
    mock_get_audio_type.assert_called_once_with(b"audiodata")



from typing import cast

from vertexai.generative_models import GenerationConfig

from ...base.call_params import CommonCallParams
from ..call_params import VertexCallParams

VERTEX_PARAM_MAPPING = {
    "temperature": "temperature",
    "max_tokens": "max_output_tokens",
    "top_p": "top_p",
    "stop": "stop_sequences",
}


def convert_common_call_params(common_params: CommonCallParams) -> VertexCallParams:
    """Convert CommonCallParams to Vertex parameters."""
    generation_config = {}

    for key, value in common_params.items():
        if key not in VERTEX_PARAM_MAPPING or value is None:
            continue

        if key == "stop":
            generation_config["stop_sequences"] = (
                [value] if isinstance(value, str) else value
            )
        else:
            generation_config[VERTEX_PARAM_MAPPING[key]] = value

    if not generation_config:
        return cast(VertexCallParams, {})

    return cast(
        VertexCallParams,
        {"generation_config": cast(GenerationConfig, generation_config)},
    )



from typing import cast

from vertexai.generative_models import Content

from mirascope.core import BaseMessageParam
from mirascope.core.base import DocumentPart, ImagePart, TextPart
from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)
from mirascope.core.base.message_param import ImageURLPart, ToolCallPart, ToolResultPart
from mirascope.core.vertex._utils import convert_message_params


def _is_image_mime(mime_type: str) -> bool:
    return mime_type in ["image/jpeg", "image/png", "image/gif", "image/webp"]


def _to_image_part(mime_type: str, data: bytes) -> ImagePart:
    if not _is_image_mime(mime_type):
        raise ValueError(
            f"Unsupported image media type: {mime_type}. "
            "Expected one of: image/jpeg, image/png, image/gif, image/webp."
        )
    return ImagePart(type="image", media_type=mime_type, image=data, detail=None)


def _to_document_part(mime_type: str, data: bytes) -> DocumentPart:
    if mime_type != "application/pdf":
        raise ValueError(
            f"Unsupported document media type: {mime_type}. "
            "Only application/pdf is supported."
        )
    return DocumentPart(type="document", media_type=mime_type, document=data)


class VertexMessageParamConverter(BaseMessageParamConverter):
    """Converts between Vertex `Content` and Mirascope `BaseMessageParam`."""

    @staticmethod
    def to_provider(message_params: list[BaseMessageParam]) -> list[Content]:
        """
        Convert from Mirascope `BaseMessageParam` to Vertex `Content`.
        """
        return convert_message_params(
            cast(list[BaseMessageParam | Content], message_params)
        )

    @staticmethod
    def from_provider(message_params: list[Content]) -> list[BaseMessageParam]:
        """
        Convert from Vertex's `Content` to Mirascope `BaseMessageParam`.
        """
        converted = []
        for message_param in message_params:
            role: str = (
                "assistant" if message_param.role == "model" else message_param.role
            )
            contents = []
            has_tool_call = False
            for part in message_param.parts:
                if part.function_response:
                    converted.append(
                        BaseMessageParam(
                            role=message_param.role,
                            content=[
                                ToolResultPart(
                                    type="tool_result",
                                    name=part.function_response.name,
                                    content=part.function_response.response["result"],  # pyright: ignore [reportReturnType, reportArgumentType]
                                    id=None,
                                    is_error=False,
                                )
                            ],
                        )
                    )
                elif part.inline_data:
                    blob = part.inline_data
                    mime = blob.mime_type
                    data = blob.data
                    if _is_image_mime(mime):
                        contents.append(_to_image_part(mime, data))
                    elif mime == "application/pdf":
                        contents.append(_to_document_part(mime, data))
                    else:
                        raise ValueError(
                            f"Unsupported inline_data mime type: {mime}. Cannot convert to BaseMessageParam."
                        )

                elif part.file_data:
                    if _is_image_mime(part.file_data.mime_type):
                        contents.append(
                            ImageURLPart(
                                type="image_url",
                                url=part.file_data.file_uri,
                                detail=None,
                            )
                        )
                    else:
                        raise ValueError(
                            f"FileData.file_uri is not support: {part.file_data}. Cannot convert to BaseMessageParam."
                        )
                elif part.function_call:
                    converted.append(
                        BaseMessageParam(
                            role=role,
                            content=[
                                ToolCallPart(
                                    type="tool_call",
                                    name=part.function_call.name,
                                    args=dict(part.function_call.args),
                                )
                            ],
                        )
                    )
                elif part.text:
                    contents.append(TextPart(type="text", text=part.text))

                else:  # pragma: no cover
                    raise ValueError(
                        "Part does not contain any supported content (text, image, or document)."
                    )

            if len(contents) == 1 and isinstance(contents[0], TextPart):
                converted.append(BaseMessageParam(role=role, content=contents[0].text))
            else:
                if contents:
                    converted.append(
                        BaseMessageParam(
                            role="tool" if has_tool_call else role, content=contents
                        )
                    )

        return converted



"""This module contains the type definition for the Google call keyword arguments."""

from collections.abc import Sequence

from google.genai.types import ContentOrDict, Tool

from ..base import BaseCallKwargs
from .call_params import GoogleCallParams


class GoogleCallKwargs(GoogleCallParams, BaseCallKwargs[Tool]):
    model: str
    contents: Sequence[ContentOrDict]



"""The `GoogleTool` class for easy tool usage with Google's Google LLM calls.

usage docs: learn/tools.md
"""

from __future__ import annotations

from typing import Any

from google.genai.types import (
    FunctionCall,
    FunctionDeclaration,
    Tool,
)
from pydantic.json_schema import SkipJsonSchema

from ..base import BaseTool


class GoogleTool(BaseTool):
    """A class for defining tools for Google LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.google import google_call


    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"


    @google_call("google-1.5-flash", tools=[format_book])
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")
    if tool := response.tool:  # returns an `GoogleTool` instance
        print(tool.call())
    ```
    """

    __provider__ = "google"

    tool_call: SkipJsonSchema[FunctionCall]

    @classmethod
    def tool_schema(cls) -> Tool:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined.

        Example:
        ```python
        from mirascope.core.google import GoogleTool


        def format_book(title: str, author: str) -> str:
            return f"{title} by {author}"


        tool_type = GoogleTool.type_from_fn(format_book)
        print(tool_type.tool_schema())  # prints the Google-specific tool schema
        ```
        """
        model_schema = cls.model_json_schema()
        fn: dict[str, Any] = {"name": cls._name(), "description": cls._description()}

        if model_schema["properties"]:
            fn["parameters"] = model_schema

        if "parameters" in fn:
            # Define a function to handle both ref resolution and examples conversion
            def resolve_refs_and_fix_examples(
                schema: dict[str, Any], defs: dict[str, Any] | None = None
            ) -> dict[str, Any]:
                """Recursively resolve $ref references and fix examples/example fields."""
                # If this is a reference, resolve it
                if "$ref" in schema:
                    ref = schema["$ref"]
                    if ref.startswith("#/$defs/") and defs:
                        ref_key = ref.replace("#/$defs/", "")
                        if ref_key in defs:
                            # Merge the definition with the current schema (excluding $ref)
                            resolved = {
                                **{k: v for k, v in schema.items() if k != "$ref"},
                                **resolve_refs_and_fix_examples(defs[ref_key], defs),
                            }
                            return resolved

                # Handle examples -> example conversion
                result = {}
                for key, value in schema.items():
                    # Convert "examples" to "example" for Google Schema
                    if key == "examples":
                        result["example"] = value
                    elif isinstance(value, dict):
                        result[key] = resolve_refs_and_fix_examples(value, defs)
                    elif isinstance(value, list):
                        result[key] = [
                            resolve_refs_and_fix_examples(item, defs)
                            if isinstance(item, dict)
                            else item
                            for item in value
                        ]
                    else:
                        result[key] = value
                return result

            # Extract $defs if they exist
            defs = {}
            if "$defs" in fn["parameters"]:
                defs = fn["parameters"].pop("$defs")

            # Resolve all references in the parameters and fix examples
            fn["parameters"] = resolve_refs_and_fix_examples(fn["parameters"], defs)

            def handle_enum_schema(prop_schema: dict[str, Any]) -> dict[str, Any]:
                if "enum" in prop_schema:
                    prop_schema["format"] = "enum"
                return prop_schema

            # Process properties after resolving references
            # We're already handling examples -> example conversion recursively above
            fn["parameters"]["properties"] = {
                prop: {
                    key: value
                    for key, value in handle_enum_schema(prop_schema).items()
                    if key != "default"
                }
                for prop, prop_schema in fn["parameters"]["properties"].items()
            }

        return Tool(function_declarations=[FunctionDeclaration(**fn)])

    @classmethod
    def from_tool_call(cls, tool_call: FunctionCall) -> GoogleTool:
        """Constructs an `GoogleTool` instance from a `tool_call`.

        Args:
            tool_call: The Google tool call from which to construct this tool instance.
        """
        model_json = {"tool_call": tool_call}
        if tool_call.args:
            model_json |= dict(tool_call.args.items())
        return cls.model_validate(model_json)



"""The `VertexTool` class for easy tool usage with Google's Vertex LLM calls.

usage docs: learn/tools.md
"""

from __future__ import annotations

from typing import Any

from google.cloud.aiplatform_v1beta1.types import FunctionCall
from pydantic.json_schema import SkipJsonSchema
from vertexai.generative_models import FunctionDeclaration, Tool

from ..base import BaseTool


class VertexTool(BaseTool):
    """A class for defining tools for Vertex LLM calls.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.vertex import vertex_call


    def format_book(title: str, author: str) -> str:
        return f"{title} by {author}"


    @vertex_call("gemini-1.5-flash", tools=[format_book])
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")
    if tool := response.tool:  # returns an `VertexTool` instance
        print(tool.call())
    ```
    """

    __provider__ = "vertex"

    tool_call: SkipJsonSchema[FunctionCall]

    @classmethod
    def tool_schema(cls) -> Tool:
        """Constructs a JSON Schema tool schema from the `BaseModel` schema defined.

        Example:
        ```python
        from mirascope.core.vertex import VertexTool


        def format_book(title: str, author: str) -> str:
            return f"{title} by {author}"


        tool_type = VertexTool.type_from_fn(format_book)
        print(tool_type.tool_schema())  # prints the Vertex-specific tool schema
        ```
        """
        model_schema = cls.model_json_schema()
        fn: dict[str, Any] = {"name": cls._name(), "description": cls._description()}
        if model_schema["properties"]:
            fn["parameters"] = model_schema
        if model_schema["required"]:
            fn["parameters"]["required"] = model_schema["required"]
        if "parameters" in fn:
            if "$defs" in fn["parameters"]:
                raise ValueError(
                    "Unfortunately Google's Vertex API cannot handle nested structures "
                    "with $defs."
                )

            def handle_enum_schema(prop_schema: dict[str, Any]) -> dict[str, Any]:
                if "enum" in prop_schema:
                    prop_schema["format"] = "enum"
                return prop_schema

            fn["parameters"]["properties"] = {
                prop: {
                    key: value
                    for key, value in handle_enum_schema(prop_schema).items()
                    if key != "default"
                }
                for prop, prop_schema in fn["parameters"]["properties"].items()
            }
        return Tool(function_declarations=[FunctionDeclaration(**fn)])

    @classmethod
    def from_tool_call(cls, tool_call: FunctionCall) -> VertexTool:
        """Constructs an `VertexTool` instance from a `tool_call`.

        Args:
            tool_call: The Vertex tool call from which to construct this tool instance.
        """
        model_json = {"tool_call": tool_call}
        if tool_call.args:
            model_json |= dict(tool_call.args.items())
        return cls.model_validate(model_json)



from collections.abc import Callable

import pytest

from mirascope.core import merge_decorators


def test_single_decorator():
    """Test basic functionality with a single decorator."""

    def capitalize_result(f: Callable[[str], str]) -> Callable[[str], str]:
        def wrapper(*args: str, **kwargs: str) -> str:
            result = f(*args, **kwargs)
            return result.upper()

        return wrapper

    merged = merge_decorators(capitalize_result)()

    @merged
    def greet(name: str) -> str:
        """Return a greeting message."""
        return f"hello {name}"

    result = greet("world")
    assert result == "HELLO WORLD"
    assert greet.__doc__ == "Return a greeting message."
    assert greet.__name__ == "greet"


def test_multiple_decorators():
    """Test with multiple decorators and verify execution order."""
    traces: list[str] = []

    def decorator_a(f: Callable) -> Callable:
        def wrapper(*args: object, **kwargs: object) -> object:
            traces.append("a")
            return f(*args, **kwargs)

        return wrapper

    def decorator_b(f: Callable) -> Callable:
        def wrapper(*args: object, **kwargs: object) -> object:
            traces.append("b")
            return f(*args, **kwargs)

        return wrapper

    def decorator_c(f: Callable) -> Callable:
        def wrapper(*args: object, **kwargs: object) -> object:
            traces.append("c")
            return f(*args, **kwargs)

        return wrapper

    merged = merge_decorators(decorator_a, decorator_b, decorator_c)()

    @merged
    def func() -> str:
        """Test function."""
        traces.append("func")
        return "result"

    result = func()
    assert result == "result"
    # Decorators are applied in the order they are passed to merge_decorators
    assert traces == ["a", "b", "c", "func"]
    assert func.__doc__ == "Test function."
    assert func.__name__ == "func"


def test_decorator_with_arguments():
    """Test decorators that take arguments."""

    def multiply(n: int) -> Callable:
        def decorator(f: Callable[[int], int]) -> Callable[[int], int]:
            def wrapper(x: int) -> int:
                return f(x) * n

            return wrapper

        return decorator

    def add(n: int) -> Callable:
        def decorator(f: Callable[[int], int]) -> Callable[[int], int]:
            def wrapper(x: int) -> int:
                return f(x) + n

            return wrapper

        return decorator

    # First multiply by 2, then add 3: (x * 2) + 3
    merged = merge_decorators(multiply(2), add(3))()

    @merged
    def calculate(x: int) -> int:
        """Calculate a value."""
        return x

    assert calculate(5) == 16
    assert calculate.__doc__ == "Calculate a value."
    assert calculate.__name__ == "calculate"


def test_nested_metadata_preservation():
    """Test that metadata is preserved through nested decorator applications."""

    def outer_decorator(text: str) -> Callable:
        def decorator(f: Callable) -> Callable:
            def wrapper(*args: object, **kwargs: object) -> object:
                return f"{text}: {f(*args, **kwargs)}"

            return wrapper

        return decorator

    def inner_decorator(f: Callable) -> Callable:
        def wrapper(*args: object, **kwargs: object) -> object:
            return f"wrapped({f(*args, **kwargs)})"

        return wrapper

    merged = merge_decorators(outer_decorator("hello"), inner_decorator)()

    @merged
    def example(x: str) -> str:
        """Example function with detailed docstring."""
        return x

    result = example("world")
    assert result == "hello: wrapped(world)"
    assert example.__doc__ == "Example function with detailed docstring."
    assert example.__name__ == "example"


def test_error_propagation():
    """Test that errors are properly propagated through the decorator chain."""

    def error_decorator(f: Callable) -> Callable:
        def wrapper(*args: object, **kwargs: object) -> None:
            raise ValueError("Test error")

        return wrapper

    merged = merge_decorators(error_decorator)()

    @merged
    def func() -> str:
        """Function that should raise an error."""
        return "success"  # pragma: no cover

    with pytest.raises(ValueError, match="Test error"):
        func()
    assert func.__doc__ == "Function that should raise an error."
    assert func.__name__ == "func"


def test_async_function_compatibility():
    """Test compatibility with async functions."""
    import asyncio

    def log_decorator(f: Callable) -> Callable:
        async def wrapper(*args: object, **kwargs: object) -> object:
            result = await f(*args, **kwargs)
            return f"logged: {result}"

        return wrapper

    merged = merge_decorators(log_decorator)()

    @merged
    async def async_func(x: str) -> str:
        """Async test function."""
        return x

    result = asyncio.run(async_func("test"))
    assert result == "logged: test"
    assert async_func.__doc__ == "Async test function."
    assert async_func.__name__ == "async_func"


def test_metadata_preservation():
    """Test that the original function's metadata is preserved."""

    def decorator(f: Callable[[int], str]) -> Callable[[int], str]:
        def wrapper(x: int) -> str:
            return f(x)

        return wrapper

    merged = merge_decorators(decorator)()

    @merged
    def func(x: int) -> str:
        """Convert int to str."""
        return str(x)

    # Original function's type hints are preserved
    from typing import get_type_hints

    type_hints = get_type_hints(func)
    assert type_hints["x"] is int
    assert type_hints["return"] is str
    assert func.__doc__ == "Convert int to str."
    assert func.__name__ == "func"
    assert func(16) == "16"



"""This module contains the type definition for the Mistral call keyword arguments."""

from collections.abc import Sequence
from typing import Any

from mistralai.models import (
    AssistantMessage,
    SystemMessage,
    ToolMessage,
    UserMessage,
)

from ..base import BaseCallKwargs
from .call_params import MistralCallParams


class MistralCallKwargs(MistralCallParams, BaseCallKwargs[dict[str, Any]]):
    model: str
    messages: Sequence[AssistantMessage | SystemMessage | ToolMessage | UserMessage]



"""Tests the `gemini._utils.convert_message_params` function."""

import io
from unittest.mock import MagicMock, patch

import pytest
from google.generativeai import protos
from google.generativeai.types import ContentDict

from mirascope.core.base import (
    AudioPart,
    AudioURLPart,
    BaseMessageParam,
    CacheControlPart,
    ImagePart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.gemini._utils._convert_message_params import convert_message_params


@patch("PIL.Image.open", new_callable=MagicMock)
def test_convert_message_params(mock_image_open: MagicMock) -> None:
    """Tests the `convert_message_params` function."""
    mock_image_open.return_value = "test"
    message_params: list[BaseMessageParam | ContentDict] = [
        BaseMessageParam(role="system", content="You are a helpful assistant."),
        BaseMessageParam(role="user", content="Hello"),
        {"role": "user", "parts": ["Hello", {"type": "text", "text": "Hello"}]},  # pyright: ignore[reportAssignmentType]
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="test"),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail=None
                ),
                AudioPart(type="audio", media_type="audio/wav", audio=b"audio"),
                ToolResultPart(
                    name="tool_name", id="tool_id", content="result", type="tool_result"
                ),
            ],
        ),
        BaseMessageParam(
            role="model",
            content=[
                TextPart(type="text", text="test"),
                ToolCallPart(type="tool_call", name="tool_name", args={"arg": "val"}),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params == [
        {"role": "system", "parts": ["You are a helpful assistant."]},
        {"role": "user", "parts": ["Hello"]},
        {"role": "user", "parts": ["Hello", {"type": "text", "text": "Hello"}]},
        {
            "role": "user",
            "parts": ["test", "test", {"mime_type": "audio/wav", "data": b"audio"}],
        },
        {
            "role": "user",
            "parts": [
                protos.FunctionResponse(
                    name="tool_name", response={"result": "result"}
                ),
            ],
        },
        {
            "role": "model",
            "parts": [
                "test",
                protos.FunctionCall(
                    name="tool_name",
                    args={"arg": "val"},
                ),
            ],
        },
    ]

    one_part = BaseMessageParam(
        role="user",
        content=[
            TextPart(type="text", text="test"),
        ],
    )
    converted_message_params = convert_message_params([one_part])
    assert converted_message_params == [{"parts": ["test"], "role": "user"}]

    mock_image_open.assert_called_once()
    bytes_io: io.BytesIO = mock_image_open.call_args.args[0]
    assert bytes_io.getvalue() == b"image"

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. Gemini currently only supports "
        "JPEG, PNG, WebP, HEIC, and HEIF images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail=None,
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/unknown. "
        "Gemini currently only supports WAV, MP3, AIFF, AAC, OGG, and FLAC audio file types.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(
                            type="audio", media_type="audio/unknown", audio=b"audio"
                        )
                    ],
                ),
            ]
        )

    with pytest.raises(
        ValueError,
        match="Gemini currently only supports text, image, and audio parts. Part provided: cache_control",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        CacheControlPart(type="cache_control", cache_type="ephemeral")
                    ],
                )
            ]
        )


def test_system_message_non_string_content() -> None:
    """Test that a system message with non-string content raises ValueError."""
    with pytest.raises(
        ValueError,
        match="System message content must be a single text string.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="system",
                    content=[TextPart(type="text", text="not a string")],
                )
            ]
        )


@patch(
    "mirascope.core.gemini._utils._convert_message_params._load_media",
    return_value=b"image_url_data",
)
@patch("PIL.Image.open", new_callable=MagicMock)
def test_image_url_with_http_valid(
    mock_image_open: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test image_url part with an HTTP URL and valid media type."""
    fake_image = MagicMock()
    fake_image.format = "JPEG"
    mock_image_open.return_value = fake_image

    message = BaseMessageParam(
        role="user",
        content=[
            ImageURLPart(type="image_url", url="https://example.com/image", detail=None)
        ],
    )
    result = convert_message_params([message])
    # Expect the image to be processed via PIL.Image.open and returned in parts.
    assert result == [{"role": "user", "parts": [fake_image]}]
    # Check that _load_media was called with the URL.
    mock_load_media.assert_called_once_with("https://example.com/image")


@patch(
    "mirascope.core.gemini._utils._convert_message_params._load_media",
    return_value=b"image_url_data",
)
@patch("PIL.Image.open", new_callable=MagicMock)
def test_image_url_with_http_invalid_media_type(
    mock_image_open: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test image_url part with an HTTP URL and an invalid media type raises ValueError."""
    # Use patch.dict to temporarily set PIL.Image.MIME for an invalid format.
    with patch.dict("PIL.Image.MIME", {"SVG": "image/svg"}, clear=False):
        fake_image = MagicMock()
        fake_image.format = "SVG"
        mock_image_open.return_value = fake_image

        message = BaseMessageParam(
            role="user",
            content=[
                ImageURLPart(
                    type="image_url", url="https://example.com/image", detail=None
                )
            ],
        )
        with pytest.raises(
            ValueError,
            match="Unsupported image media type: image/svg. Gemini currently only supports JPEG, PNG, WebP, HEIC, and HEIF images.",
        ):
            convert_message_params([message])


@patch("mirascope.core.gemini._utils._convert_message_params._load_media")
def test_image_url_with_non_http(mock_load_media: MagicMock) -> None:
    """Test image_url part with a non-HTTP URL returns a FileData object."""
    # For non-HTTP URLs, _load_media should not be called.
    message = BaseMessageParam(
        role="user",
        content=[
            ImageURLPart(type="image_url", url="file://local/path/image", detail=None)
        ],
    )
    result = convert_message_params([message])
    # Expect the part to be a FileData with file_uri equal to the URL.
    assert len(result) == 1
    part = result[0]["parts"][0]
    assert isinstance(part, protos.FileData)
    assert part.file_uri == "file://local/path/image"
    mock_load_media.assert_not_called()


@patch(
    "mirascope.core.gemini._utils._convert_message_params._load_media",
    return_value=b"audio_data",
)
@patch(
    "mirascope.core.gemini._utils._convert_message_params.get_audio_type",
    return_value="wav",
)
def test_audio_url_with_http_valid(
    mock_get_audio_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test audio_url part with an HTTP URL and a valid audio type."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="https://example.com/audio")],
    )
    result = convert_message_params([message])
    assert result == [
        {"role": "user", "parts": [{"mime_type": "audio/wav", "data": b"audio_data"}]}
    ]
    mock_load_media.assert_called_once_with("https://example.com/audio")
    mock_get_audio_type.assert_called_once_with(b"audio_data")


@patch(
    "mirascope.core.gemini._utils._convert_message_params._load_media",
    return_value=b"audio_data",
)
@patch(
    "mirascope.core.gemini._utils._convert_message_params.get_audio_type",
    return_value="unknown",
)
def test_audio_url_with_http_invalid(
    mock_get_audio_type: MagicMock, mock_load_media: MagicMock
) -> None:
    """Test audio_url part with an HTTP URL and an invalid audio type raises ValueError."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="https://example.com/audio")],
    )
    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/unknown. Gemini currently only supports WAV, MP3, AIFF, AAC, OGG, and FLAC audio file types.",
    ):
        convert_message_params([message])
    mock_load_media.assert_called_once_with("https://example.com/audio")
    mock_get_audio_type.assert_called_once_with(b"audio_data")


def test_audio_url_with_non_http() -> None:
    """Test audio_url part with a non-HTTP URL returns a FileData object."""
    message = BaseMessageParam(
        role="user",
        content=[AudioURLPart(type="audio_url", url="ftp://example.com/audio")],
    )
    result = convert_message_params([message])
    assert len(result) == 1
    part = result[0]["parts"][0]
    assert isinstance(part, protos.FileData)
    assert part.file_uri == "ftp://example.com/audio"


def test_tool_result_without_prior_parts() -> None:
    """Test tool_result part when there are no preceding parts."""
    message = BaseMessageParam(
        role="user",
        content=[
            ToolResultPart(type="tool_result", name="tool_only", content="only result")
        ],
    )
    result = convert_message_params([message])
    # Expect a single message with role "user" and a FunctionResponse part.
    assert result == [
        {
            "role": "user",
            "parts": [
                protos.FunctionResponse(
                    name="tool_only", response={"result": "only result"}
                )
            ],
        }
    ]


def test_string_content_non_user() -> None:
    """Test that a message with string content for a non-user role converts to model role."""
    message = BaseMessageParam(role="assistant", content="Hi there")
    result = convert_message_params([message])
    # For non-user roles, the role is set to "model"
    assert result == [{"role": "model", "parts": ["Hi there"]}]



"""This module contains the Stream class that inherits from BaseStream."""

from __future__ import annotations

from collections.abc import AsyncGenerator, Generator
from typing import Any, TypeVar

from ..core.base import (
    BaseCallParams,
    BaseCallResponse,
    BaseCallResponseChunk,
    BaseDynamicConfig,
    BaseMessageParam,
    BaseTool,
)
from ..core.base.call_response import JsonableType
from ..core.base.stream import BaseStream
from ..core.base.types import FinishReason
from ..llm.call_response import CallResponse
from ..llm.call_response_chunk import CallResponseChunk
from .tool import Tool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_UserMessageParamT = TypeVar("_UserMessageParamT", bound=BaseMessageParam)
_AssistantMessageParamT = TypeVar("_AssistantMessageParamT", bound=BaseMessageParam)
_ToolMessageParamT = TypeVar("_ToolMessageParamT", bound=BaseMessageParam)
_MessageParamT = TypeVar("_MessageParamT", bound=BaseMessageParam)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ToolSchemaT = TypeVar("_ToolSchemaT")
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_FinishReasonT = TypeVar("_FinishReasonT")


class Stream(
    BaseStream[
        BaseCallResponse,
        BaseCallResponseChunk,
        BaseMessageParam,
        BaseMessageParam,
        BaseMessageParam,
        BaseMessageParam,
        BaseTool,
        Any,
        BaseDynamicConfig,
        BaseCallParams,
        FinishReason,
    ],
):
    """A non-pydantic class that inherits from BaseStream."""

    _stream: BaseStream[
        BaseCallResponse,
        BaseCallResponseChunk,
        BaseMessageParam,
        BaseMessageParam,
        BaseMessageParam,
        BaseMessageParam,
        BaseTool,
        Any,
        BaseDynamicConfig,
        BaseCallParams,
        FinishReason,
    ]

    def __init__(
        self,
        *,
        stream: BaseStream[
            BaseCallResponse,
            BaseCallResponseChunk,
            BaseMessageParam,
            BaseMessageParam,
            BaseMessageParam,
            BaseMessageParam,
            BaseTool,
            Any,
            BaseDynamicConfig,
            BaseCallParams,
            FinishReason,
        ],
    ) -> None:
        """Initialize the Stream class."""
        object.__setattr__(self, "_stream", stream)

    def __getattribute__(self, name: str) -> Any:  # noqa: ANN401
        special_names = {
            "_stream",
            "cost",
            "_construct_message_param",
            "construct_call_response",
            "tool_message_params",
            "__dict__",
            "__class__",
            "__repr__",
            "__str__",
            "__iter__",
            "__aiter__",
        }

        if name in special_names:
            return object.__getattribute__(self, name)
        response = object.__getattribute__(self, "_stream")
        return getattr(response, name)

    def __iter__(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
    ) -> Generator[
        tuple[CallResponseChunk, Tool[_ToolMessageParamT] | None],
        None,
        None,
    ]:
        """Iterate over the stream."""
        for chunk, tool in self._stream:
            yield (
                CallResponseChunk(response=chunk),  # pyright: ignore [reportAbstractUsage]
                Tool(tool=tool) if tool is not None else None,  # pyright: ignore [reportAbstractUsage]
            )

    async def __aiter__(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
    ) -> AsyncGenerator[
        tuple[CallResponseChunk, Tool[_ToolMessageParamT] | None],
        None,
    ]:
        """Iterates over the stream and stores useful information."""
        async for chunk, tool in self._stream:
            yield (
                CallResponseChunk(response=chunk),  # pyright: ignore [reportAbstractUsage]
                Tool(tool=tool) if tool is not None else None,  # pyright: ignore [reportAbstractUsage]
            )

    @property
    def cost(self) -> float | None:
        return self._stream.cost

    def _construct_message_param(
        self, tool_calls: list[Any] | None = None, content: str | None = None
    ) -> BaseMessageParam:
        return self._stream._construct_message_param(
            tool_calls=tool_calls, content=content
        )

    def construct_call_response(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
    ) -> CallResponse:
        return CallResponse(response=self._stream.construct_call_response())  # pyright: ignore [reportAbstractUsage]

    @classmethod
    def tool_message_params(  # pyright: ignore [reportIncompatibleMethodOverride]
        cls, tools_and_outputs: list[tuple[Tool, JsonableType]]
    ) -> list[BaseMessageParam]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The list of tools and their outputs from which the tool
                message parameters should be constructed.
        """
        return CallResponse.tool_message_params(tools_and_outputs)



from typing import cast

from google.genai import Client
from google.genai.types import (
    Content,
    ContentDict,
    ContentOrDict,
)

from mirascope.core import BaseMessageParam
from mirascope.core.base import DocumentPart, DocumentURLPart, ImagePart, TextPart
from mirascope.core.base._utils._base_message_param_converter import (
    BaseMessageParamConverter,
)
from mirascope.core.base.message_param import (
    AudioPart,
    AudioURLPart,
    ImageURLPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.google._utils import convert_message_params

from ._validate_media_type import (
    _check_audio_media_type,
    _check_document_media_type,
    _check_image_media_type,
)


class GoogleMessageParamConverter(BaseMessageParamConverter):
    """Converts between Google `ContentDict` and Mirascope `BaseMessageParam`."""

    @staticmethod
    def to_provider(
        message_params: list[BaseMessageParam], client: Client | None = None
    ) -> list[ContentDict]:
        """
        Convert from Mirascope `BaseMessageParam` to Google `ContentDict`.
        """
        return convert_message_params(
            cast(list[BaseMessageParam | ContentDict], message_params),
            client or Client(),
        )

    @staticmethod
    def from_provider(message_params: list[ContentOrDict]) -> list[BaseMessageParam]:
        """
        Convert from Google's `ContentDict` to Mirascope `BaseMessageParam`.
        """
        converted: list[BaseMessageParam] = []
        for message_param in message_params:
            if isinstance(message_param, dict):
                message_param = Content.model_validate(message_param)
            role = message_param.role
            if not role or role == "model":
                role = "assistant"
            content_list = []
            for part in message_param.parts or []:
                if part.text:
                    content_list.append(TextPart(type="text", text=part.text))

                elif blob := part.inline_data:
                    mime_type = blob.mime_type or ""
                    data = blob.data or b""
                    if mime_type.startswith("image/"):
                        _check_image_media_type(mime_type)
                        content_list.append(
                            ImagePart(
                                type="image",
                                media_type=mime_type,
                                image=data,
                                detail=None,
                            )
                        )
                    elif mime_type.startswith("audio/"):
                        _check_audio_media_type(mime_type)
                        content_list.append(
                            AudioPart(
                                type="audio",
                                media_type=mime_type,
                                audio=data,
                            )
                        )
                    elif mime_type in [
                        "application/pdf",
                        "application/x-javascript",
                        "text/javascript",
                        "application/x-python",
                        "text/x-python",
                        "text/plain",
                        "text/html",
                        "text/css",
                        "text/csv",
                        "text/xml",
                        "text/rtf",
                        "text/md",
                    ]:
                        _check_document_media_type(mime_type)
                        content_list.append(
                            DocumentPart(
                                type="document", media_type=mime_type, document=data
                            )
                        )
                    else:
                        raise ValueError(
                            f"Unsupported inline_data mime type: {mime_type}. Cannot convert to BaseMessageParam."
                        )

                elif file_data := part.file_data:
                    mime_type = file_data.mime_type or ""
                    if mime_type.startswith("image/"):
                        content_list.append(
                            ImageURLPart(
                                type="image_url",
                                url=cast(str, part.file_data.file_uri),
                                detail=None,
                            )
                        )
                    elif mime_type.startswith("audio/"):
                        content_list.append(
                            AudioURLPart(
                                type="audio_url",
                                url=cast(str, part.file_data.file_uri),
                            )
                        )
                    elif mime_type in [
                        "application/pdf",
                        "application/x-javascript",
                        "text/javascript",
                        "application/x-python",
                        "text/x-python",
                        "text/plain",
                        "text/html",
                        "text/css",
                        "text/csv",
                        "text/xml",
                        "text/rtf",
                        "text/md",
                    ]:
                        content_list.append(
                            DocumentURLPart(
                                type="document_url",
                                url=cast(str, part.file_data.file_uri),
                            )
                        )
                    else:
                        # Since `FileDataDict` handles any file data, we use
                        # `ImageURLPart` for unknown mime types
                        content_list.append(
                            ImageURLPart(
                                type="image_url",
                                url=cast(str, part.file_data.file_uri),
                                detail=None,
                            )
                        )
                elif part.function_call:
                    converted.append(
                        BaseMessageParam(
                            role=role,
                            content=[
                                ToolCallPart(
                                    type="tool_call",
                                    name=part.function_call.name,  # pyright: ignore [reportArgumentType]
                                    args=part.function_call.args,
                                )
                            ],
                        )
                    )
                elif part.function_response:
                    converted.append(
                        BaseMessageParam(
                            role=role,
                            content=[
                                ToolResultPart(
                                    type="tool_result",
                                    name=part.function_response.name,  # pyright: ignore [reportArgumentType]
                                    content=part.function_response.response[  # pyright: ignore [reportArgumentType, reportOptionalSubscript]
                                        "content"
                                    ],
                                    id=None,
                                    is_error=False,
                                )
                            ],
                        )
                    )
                else:
                    raise ValueError(
                        "Part does not contain any supported content (text, image, audio, or document)."
                    )

            if len(content_list) == 1 and isinstance(content_list[0], TextPart):
                converted.append(
                    BaseMessageParam(role=role, content=content_list[0].text)
                )
            else:
                if content_list:
                    converted.append(BaseMessageParam(role=role, content=content_list))

        return converted



"""Tests the `mistral.call_response` module."""

from mistralai.models import (
    AssistantMessage,
    ChatCompletionChoice,
    ChatCompletionResponse,
    FunctionCall,
    ToolCall,
    ToolMessage,
    UsageInfo,
    UserMessage,
)

from mirascope.core import BaseMessageParam
from mirascope.core.mistral.call_response import MistralCallResponse
from mirascope.core.mistral.tool import MistralTool


def test_mistral_call_response() -> None:
    """Tests the `MistralCallResponse` class."""
    choices = [
        ChatCompletionChoice(
            index=0,
            message=AssistantMessage(content="content"),
            finish_reason="stop",
        )
    ]
    usage = UsageInfo(prompt_tokens=1, completion_tokens=1, total_tokens=2)
    completion = ChatCompletionResponse(
        id="id",
        choices=choices,
        created=0,
        model="mistral-large-latest",
        object="",
        usage=usage,
    )
    call_response = MistralCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response._provider == "mistral"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["stop"]
    assert call_response.model == "mistral-large-latest"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost == 8e-06
    assert call_response.message_param == AssistantMessage(content="content")
    assert call_response.tools is None
    assert call_response.tool is None
    assert call_response.common_finish_reasons == ["stop"]
    assert call_response.common_message_param == BaseMessageParam(
        role="assistant", content="content"
    )
    assert call_response.common_user_message_param is None
    call_response.user_message_param = UserMessage(content="content")
    assert call_response.common_user_message_param == BaseMessageParam(
        role="user", content="content"
    )


def test_mistral_call_response_with_tools() -> None:
    """Tests the `MistralCallResponse` class with tools."""

    class FormatBook(MistralTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ToolCall(
        id="id",
        function=FunctionCall(
            name="FormatBook",
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
        ),
        type="function",
    )
    completion = ChatCompletionResponse(
        id="id",
        choices=[
            ChatCompletionChoice(
                finish_reason="stop",
                index=0,
                message=AssistantMessage(content="content", tool_calls=[tool_call]),
            )
        ],
        created=0,
        model="mistral-large-latest",
        object="",
        usage=UsageInfo(prompt_tokens=1, completion_tokens=1, total_tokens=2),
    )
    call_response = MistralCallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    assert call_response.tool_message_params([(tool, output)]) == [
        ToolMessage(
            role="tool",
            content=output,
            tool_call_id=tool_call.id,
            name="FormatBook",  # type: ignore
        )
    ]



from mirascope.core.base._utils._convert_provider_finish_reason_to_finish_reason import (
    FinishReasonMappingValue,
    _convert_finish_reasons_to_common_finish_reasons_from_mapping,
)
from mirascope.core.base.types import FinishReason

_FinishReasonMapping: dict[str, FinishReasonMappingValue] = {
    "end_turn": "stop",
    "max_tokens": "length",
    "stop_sequence": "stop",
    "tool_use": "tool_calls",
}


def _convert_finish_reasons_to_common_finish_reasons(
    finish_reasons: list[str],
) -> list[FinishReason] | None:
    """Provider-agnostic finish reasons."""
    return _convert_finish_reasons_to_common_finish_reasons_from_mapping(
        finish_reasons, _FinishReasonMapping
    )



"""This module defines the function return type for functions as LLM calls."""

from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI
from openai.types.chat import ChatCompletionMessageParam

from ..base import BaseDynamicConfig, BaseMessageParam
from .call_params import OpenAICallParams

AsyncOpenAIDynamicConfig = BaseDynamicConfig[
    ChatCompletionMessageParam | BaseMessageParam,
    OpenAICallParams,
    AsyncOpenAI | AsyncAzureOpenAI,
]

OpenAIDynamicConfig = BaseDynamicConfig[
    ChatCompletionMessageParam | BaseMessageParam,
    OpenAICallParams,
    OpenAI | AzureOpenAI,
]
"""The function return type for functions wrapped with the `openai_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.openai import OpenAIDynamicConfig, openai_call


@openai_call("gpt-4o-mini")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> OpenAIDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



"""Tests the `_utils.get_document_type` module."""

import pytest

from mirascope.core.base._utils._get_document_type import get_document_type

PDF_DATA = b"%PDF-1.0\ntrailer<</Root<</Pages<</Kids[]>>>>>>%%EOF"


def test_get_document_type() -> None:
    """Test the get_image_type function."""
    assert get_document_type(PDF_DATA) == "pdf"
    with pytest.raises(ValueError, match="Unsupported document type"):
        get_document_type(b"invalid")



"""This module defines the function return type for functions as LLM calls."""

from typing import TypeAlias

from ..openai import AsyncOpenAIDynamicConfig, OpenAIDynamicConfig

AsyncLiteLLMDynamicConfig: TypeAlias = AsyncOpenAIDynamicConfig
LiteLLMDynamicConfig: TypeAlias = OpenAIDynamicConfig
